"""Implementation of Optimize Base Class.

Support Optimization using gurobipy and COBRApy interfaces.

Load SBML coded metabolic model via sbmlxdf
in case of gurobipy mode: Configure gp model:
    - variables
    - constraints
    - objective function

collect data structures required during optimization and results processing

Peter Schubert, HHU Duesseldorf, CCB, June 2024
"""

import re
import os
import time
import numpy as np
import pandas as pd
from collections import defaultdict
import sbmlxdf
import f2xba.prefixes as pf
from f2xba.utils.mapping_utils import get_srefs, parse_reaction_string, valid_sbml_sid

# gurobipy should not be a hard requirement, unless used in this context
try:
    import gurobipy as gp
except ImportError:
    gp = None
    pass

XML_SPECIES_NS = 'http://www.hhu.de/ccb/rba/species/ns'
XML_COMPARTMENT_NS = 'http://www.hhu.de/ccb/rba/compartment/ns'

status2text = {1: 'LOADED', 2: 'OPTIMAL', 3: 'INFEASIBLE', 4: 'INF_OR_UNBD', 5: 'UNBOUNDED',
               6: 'CUTOFF', 7: 'ITERATION_LIMIT', 8: 'NODE_LIMIT', 9: 'TIME_LIMIT',
               10: 'SOLUTION_LIMIT', 11: 'INTERRUPTED', 12: 'NUMERIC', 13: 'SUBOPTIMAL',
               14: 'INPROGRESS', 15: 'USER_OBJ_LIMIT', 16: 'WORK_LIMIT', 17: 'MEM_LIMIT'}

def extract_net_fluxes(fluxes):
    """Extract net fluxes reactions fluxes for optimization solution.

    In extended models, reactions get split per direction and isoenzyme.
    Here, the fluxes are collected on the level of the original net reaction.

    :param fluxes: reaction fluxes from optimization solution
    :type fluxes: pandas.Series or None
    :return: net fluxes, by combining fluxes of related reactions.
    :rtype: pandas.Series or None
    """
    if fluxes is not None:
        net_fluxes = defaultdict(float)
        for rid, flux in fluxes.items():
            rid = str(rid)
            if re.match('.*_REV$', rid):
                net_fluxes[re.sub('_REV$', '', rid)] -= flux
            else:
                net_fluxes[rid] += flux
        return pd.Series(net_fluxes, name='fluxes')
    else:
        return fluxes


class Solution:
    """Structure holding information from optimization result.

    Aligned to COBRApy cobra.Solution.

    Example: Load and optimize a GECKO model using the gurobipy interface.

    .. code-block:: python

        from f2xba import EcmOptimization

        eo = EcmOptimization('iML1515_GECKO.xml')
        eo.medium = {rid: 1000.0 for rid in lb_medium}
        solution = eo.optimize()
        solution.objective_value

    Code for __repr__(), _repr_html_() copied from cobra.Solution.
    """


    def __init__(self, status, objective_value=np.nan, fluxes=None, reduced_costs=None, shadow_prices=None):

        self.status = status
        """Status string from optimization."""

        self.objective_value = objective_value
        """Objective value (float)."""

        self.fluxes = fluxes
        """Values of optimization variables (pandas Series)."""

        self.reduced_costs = reduced_costs
        """Reduced cost of optimization variables (pandas Series)."""

        self.shadow_prices = shadow_prices
        """Shadow prices of constraints (pandas Series)."""

    def __repr__(self):
        if self.status != 'optimal':
            return f'<Solution {self.status} at {id(self):#x}>'
        return f'<Solution {self.objective_value:.3f} at {id(self):#x}>'

    def _repr_html_(self):
        if self.status == 'optimal':
            with pd.option_context('display.max_rows', 10):
                df = pd.DataFrame({'fluxes': self.fluxes, 'reduced_costs': self.reduced_costs})
                html = ('<strong><em>Optimal</em> solution with objective '
                        f'value {self.objective_value:.3f}</strong><br>{df._repr_html_()}')
        else:
            html = f'<strong><em>{self.status}</em> solution</strong>'
        return html

class Optimize:
    """Support optimization of SBML encoded extended models generated by f2xba.

    `Optimize` is parent to `FbaOptimization`, `EcmOptimization`
    and `RbaOptimization` classes. `Optimize` provides common functionality, including
    support for thermodynamics constraint model variants.

    Both COBRApy and gurobipy interfaces are supported. Use of gurobipy interface
    requires GUROBI and gurobipy being installed on your system. The gurobipy interface
    has significant speed advantages, especially for larger amd more complex models.

    Example 1: Load and optimize GECKO model using the COBRApy interface.

    .. code-block:: python

        import cobra
        from f2xba import EcmOptimization

        ecm = cobra.io.read_sbml_model('iML1515_GECKO.xml')
        eo = EcmOptimization('iML1515_GECKO.xml', ecm)

        ecm.medium = {rid: 1000.0 for rid in medium}
        solution = ecm.optimize()

    Example 2: Load and optimize GECKO model using the gurobipy interface.

    .. code-block:: python

        from f2xba import EcmOptimization

        eo = EcmOptimization('iML1515_GECKO.xml')
        eo.medium = {rid: 1000.0 for rid in medium}
        solution = eo.optimize()

    """

    def __init__(self, model_type, fname, cobra_model=None):
        """Instantiate the Optimize base class

        This instantiation is invoked by child classes.
        The SBML-coded extended model is loaded from file, and information is extracted.
        A gurobipy model (gpm) is created, if gurobipy is available and no COBRApy model is provided.
        For TD constraint models, thermodynamics related variables and constraints will be reconfigured.

        :param str model_type: base type of the model ('FBA', 'ECM' or 'RBA')
        :param str fname: filename of the SBML coded extended model
        :param cobra_model: reference to cobra model (default: None)
        :type cobra_model: :class:`cobra.Model`
        """
        self.model_type = model_type
        if not os.path.exists(fname):
            print(f'Error: {fname} not found!')
            raise FileNotFoundError

        self.model_name = os.path.basename(fname).split('.')[0]
        # load SBML coded metabolic model into sbmlxdf for data extraction
        sbml_model = sbmlxdf.Model(fname)
        print(f'SBML model loaded by sbmlxdf: {fname} ({time.ctime(os.path.getmtime(fname))})')

        self.m_dict = sbml_model.to_df()
        """Information extracted from the SBML encoded model."""

        self.gpm = None
        """Instance of gurobipy.Model, in case gurobipy interface is used."""

        self.orig_gpm = None
        self.gp_neg_var = {}

        sbml_parameters = self.m_dict['parameters']['value'].to_dict()
        self.avg_enz_saturation = sbml_parameters.get('avg_enz_sat')
        """Average enzyme saturation level configured in ECM and RBA models."""

        self.protein_per_gdw = sbml_parameters.get('gP_per_gDW')
        """Total protein per gram cell dry weight (gP/gDW) configured in ECM models."""

        self.modeled_protein_mf = sbml_parameters.get('gmodeledP_per_gP')
        """Protein mass fraction explicitly modelled in the ECM."""

        self.importer_km = sbml_parameters.get('default_importer_km_value')
        """Default Michaelis Menten constant KM (mmol/l) configured in the RBA model."""

        if cobra_model:
            self.is_gpm = False
            self.model = cobra_model
            self._cp_configure_td_model_constraints()
        else:
            self.is_gpm = True
            self.var_id2gpm = {}
            self.constr_id2gpm = {}
            self.gpm = self._gp_create_model()
        self.report_model_size()

        self.locus2uid = self._get_locus2uid()
        """Map gene identifier to UniProt protein identifier."""

        self.mid2name = {re.sub(f'^{pf.M_}', '', sid): row['name']
                         for sid, row in self.m_dict['species'].iterrows()}

        self.id2groups = self._get_id2groups()
        """Map reaction identifier to group names configured in the SBML model."""

        self.rdata_model = self._get_reaction_data()
        """Information related to reactions."""

        self.uptake_rids = [rid for rid, rdata in self.rdata_model.items() if rdata['is_exchange'] is True]
        """Exchange reactions used for exchange of metabolites with the environment."""

        self.ex_rid2sid = self._get_ex_rid2sid()
        self.sidx2ex_ridx = {re.sub(f'^{pf.M_}', '', sid): re.sub(f'^{pf.R_}', '', rid)
                             for rid, sid in self.ex_rid2sid.items()}

        # rdata and net_rdata used in results processing (COBRApy ids, without 'R_', 'M_'
        self.rdata = {}
        """Information related to reactions (prefixes `R_` and `M_` removed)."""

        for rid, record in self.rdata_model.items():
            cp_rdata = record.copy()
            cp_rdata['net_rid'] = re.sub(f'^{pf.R_}', '', record['net_rid'])
            cp_rdata['reaction_str'] = re.sub(r'\b' + f'{pf.M_}', '', record['reaction_str'])
            self.rdata[re.sub(f'^{pf.R_}', '', rid)] = cp_rdata

        self.net_rdata = self._extract_net_reaction_data(self.rdata)
        """Information related to net reactions (iso-reactions and fwd/rev directions combined)."""

        self.rids_catalyzed = self.get_rids_catalyzed()
        """Map reaction identifier to related genes identifiers."""

        # self.uid2locus = {uid: locus for locus, uid in self.locus2uid.items()}

    def _get_id2groups(self):
        """Extract Groups component information from SBML model to map ids to group names.

        Optional SBML Groups package may contain assignment of reaction ids to specific groups.

        :return: mapping of model identifies to group names as per GROUPS component
        :rtype: dict (key: model id / str, val: list if group names / str)
        """
        id2groups = defaultdict(list)
        if 'groups' in self.m_dict:
            for _, row in self.m_dict['groups'].iterrows():
                members_list = row.get('members')
                if type(members_list) is str:
                    members_list = row.get('members', '')
                    for item in members_list.split(';'):
                        if '=' in item:
                            reference_id = item.split('=')[1].strip()
                            id2groups[reference_id].append(row['name'])
        return dict(id2groups)

    def _get_locus2uid(self):
        """Extract mapping for gene locus to protein id from SBML model.

        For EC models (e.g. GECKO) extract mapping
        - from protein concentration variables 'V_PC_xxx', which carry the protein name in the id
        - from fbcGeneProdAssoc of such variable extract the (single) gene product id
        - from fbcGeneProducts component extract the related 'label' - gene locus.
        for RBA models extract mapping from
        - from macromolecular coupling constraints 'MM_prot_xxx, which carry the gene locus in the name
        - from miriamAnnotation extract protein id, if available (not available for e.g. RNA or dummy protein)

        :return: mapping of gene locus to protein identifier
        :rtype: dict (key: gene locus ('label' of fbaGeneProducts), val: protein id)
        """
        locus2uid = {}
        # support for EC based models, e.g. GECKO
        for varid, row in self.m_dict['reactions'].iterrows():
            if re.match(f'{pf.V_PC_}', varid):
                uid = re.sub(f'{pf.V_PC_}', '', varid)
                gpr = row['fbcGeneProdAssoc']
                if type(gpr) is str:
                    gpid = gpr.split('=')[1]
                    label = valid_sbml_sid(self.m_dict['fbcGeneProducts'].at[gpid, 'label'])
                    locus2uid[label] = uid
        # support for RBA based models
        for constr_id, row in self.m_dict['species'].iterrows():
            if re.match(f'{pf.MM_}', constr_id):
                uids = sbmlxdf.misc.get_miriam_refs(row['miriamAnnotation'], 'uniprot')
                if len(uids) == 1:
                    label = re.sub(f'{pf.MM_}', '', constr_id)
                    locus2uid[label] = uids[0]
        return locus2uid

    def _get_ex_rid2sid(self):
        """Get mapping of cobra exchange reaction ids to RBA uptake metabolites

        From XML annotation of compartments identify the compartent for media uptake
        Note: RBA allows other compartments than external compartment to
        be used in Michaelis-Menten saturation terms for medium uptake,
        e.g. periplasm (however, this may block reactions in periplasm)

        replace compartment postfix of exchanged metabolited by medium cid of RBA model.

        Create mapping table from exchange reaction to medium
        """
        medium_cid = 'e'
        for cid, row in self.m_dict['compartments'].iterrows():
            xml_annots = row.get('xmlAnnotation')
            xml_attrs = sbmlxdf.misc.extract_xml_attrs(xml_annots, ns=XML_COMPARTMENT_NS)
            if 'medium' in xml_attrs and xml_attrs['medium'].lower == 'true':
                medium_cid = cid
                break

        ex_rid2sid = {}
        for uptake_rid in self.uptake_rids:
            row = self.m_dict['reactions'].loc[uptake_rid]
            # ridx = re.sub(f'^{pf.R_}', '', uptake_rid)
            sref_str = row['reactants'].split(';')[0]
            sid = sbmlxdf.extract_params(sref_str)['species']
            mid = sid.rsplit('_', 1)[0]
            ex_rid2sid[uptake_rid] = f'{mid}_{medium_cid}'
        return ex_rid2sid

    def report_model_size(self):
        """Print information about the model size.

        .. code-block:: python

            eo.report_model_size()
        """
        n_vars_drg = 0
        n_vars_ec = 0
        n_vars_pmc = 0
        if self.is_gpm:
            self.gpm.update()
            for var in self.gpm.getVars():
                if re.match(pf.V_DRG_, var.VarName):
                    n_vars_drg += 1
                elif re.match(pf.V_EC_, var.VarName):
                    n_vars_ec += 1
                elif re.match(pf.V_PMC_, var.VarName):
                    n_vars_pmc += 1
            model_type = 'MILP' if self.gpm.ismip else 'LP'
            print(f'{model_type} Model of {self.gpm.modelname}')
            print(f'{self.gpm.numvars} variables, {self.gpm.numconstrs} constraints, '
                  f'{self.gpm.numnzs} non-zero matrix coefficients')
        else:
            for rxn in self.model.reactions:
                if re.match(pf.V_DRG_, rxn.id):
                    n_vars_drg += 1
                elif re.match(pf.V_EC_, rxn.id):
                    n_vars_ec += 1
                elif re.match(pf.V_PMC_, rxn.id):
                    n_vars_pmc += 1

        build_str = []
        if n_vars_ec > 0:
            build_str.append(f'{n_vars_ec} enzymes')
        if n_vars_pmc > 0:
            build_str.append(f'{n_vars_pmc} process machines')
        if n_vars_drg > 0:
            build_str.append(f'{n_vars_drg} TD reaction constraints')
        if len(build_str) > 0:
            print(', '.join(build_str))

    # COBRAPY MODEL RELATED
    def _cp_configure_td_model_constraints(self):
        """Re-configure thermodynamics related variables and constraints in COBRApy model.
        """
        is_td = False
        for var in self.model.variables:
            for vprefix in [pf.V_FU_, pf.V_RU_]:
                if re.match(vprefix, var.name) and 'reverse' not in var.name:
                    is_td = True
                    var.type = 'binary'
        for constr in self.model.constraints:
            for cprefix in [pf.C_FFC_, pf.C_FRC_, pf.C_GFC_, pf.C_GRC_, pf.C_SU_]:
                if re.match(cprefix, constr.name):
                    constr.lb = None

        if is_td:
            print(f'Thermodynamic use variables (V_FU_xxx and V_RU_xxx) as binary')
            print(f'Thermodynamic constraints (C_F[FR]C_xxx, C_G[FR]C_xxx, C_SU_xxx) ≤ 0')

    # GUROBIPY MODEL CONSTRUCTION RELATED

    def _gp_add_add_linear_constraints(self, gpm, constrs):
        """Add linear constraints to the GurobiPy model.

        Supporting thermodynamics modeling related inequality constraints.

        :param gpm: GurobiPy Model instance
        :type gpm: :class:`gurobipy.Model`
        :param constrs:
        :return:
        """
        td_constr_prefixes = f'({pf.C_FFC_}|{pf.C_FRC_}|{pf.C_GFC_}|{pf.C_GRC_}|{pf.C_SU_})'
        self.constr_id2gpm = {}
        for constr_id, constr in constrs.items():
            if re.match(td_constr_prefixes, constr_id):
                self.constr_id2gpm[constr_id] = gpm.addLConstr(constr, '<', 0.0, constr_id)
            else:
                self.constr_id2gpm[constr_id] = gpm.addLConstr(constr, '=', 0.0, constr_id)

    def _gp_add_optimization_objective(self, gpm):
        """Add optimization objective to GurobiPy model.

        Using objective configured in the SBML model.

        :param gpm: GurobiPy Model instance
        :type gpm: :class:`gurobipy.Model`
        """
        # configure optimization objective
        df_fbc_objs = self.m_dict['fbcObjectives']
        active_odata = df_fbc_objs[df_fbc_objs['active'] == bool(True)].iloc[0]
        sense = gp.GRB.MAXIMIZE if 'max' in active_odata['type'].lower() else gp.GRB.MINIMIZE
        vars_gpm = []
        coeffs = []
        for sref_str in sbmlxdf.record_generator(active_odata['fluxObjectives']):
            params = sbmlxdf.extract_params(sref_str)
            vars_gpm.append(self.var_id2gpm[params['reac']])
            coeffs.append(float(params['coef']))
        gpm.setObjective(gp.LinExpr(coeffs, vars_gpm), sense)

    def _gp_create_model_neg_vars(self):
        """Create and configure a gurobipy model with data from metabolic model.

        selected variables get converted to non-negative variables (Not used for now)
        - however impact on RBA, ECM, pfba, fba, gp_moma and gp_room yet to be checked

        - create empty gp.model:
        - add variables (using reaction data from metabolic model)
        - construct and implement linear constraints based on reaction string
        - construct and implement linear objective
        - configure TD binary variables and TD related inequality constraints
        """
        # create empty gurobipy model
        gpm = gp.Model(self.m_dict['modelAttrs'].id)

        # set model parameters
        gpm.setParam('OutputFlag', 0)
        # COBRApy default Feasibility
        gpm.params.FeasibilityTol = 1e-7  # 1e-6 default
        gpm.params.OptimalityTol = 1e-7  # 1e-6 default
        gpm.params.IntFeasTol = 1e-7  # 1e-5 default

        # add variables to the gurobi model, with suppport of binary variables
        td_binary_var_prefixes = f'({pf.V_FU_}|{pf.V_RU_})'
        td_continuous_var_prefixes = f'({pf.V_DRG_}|{pf.V_DRG0_}|{pf.R_})'
        self.var_id2gpm = {}
        for var_id, row in self.m_dict['reactions'].iterrows():
            # configure TD forward/reverse use variables as binary
            if re.match(td_binary_var_prefixes, var_id):
                self.var_id2gpm[var_id] = gpm.addVar(lb=0.0, ub=1.0, vtype='B', name=var_id)
            # make selected continuous variables non-negative by splitting
            elif re.match(td_continuous_var_prefixes, var_id):
                neg_var_id = f'{var_id}_reverse'
                self.gp_neg_var[var_id] = neg_var_id
                self.var_id2gpm[var_id] = gpm.addVar(lb=max(0.0, row['fbcLb']),
                                                     ub=max(0.0, row['fbcUb']), vtype='C', name=var_id)
                self.var_id2gpm[neg_var_id] = gpm.addVar(lb=max(0.0, -row['fbcUb']),
                                                         ub=max(0.0, -row['fbcLb']), vtype='C', name=neg_var_id)
            # balance continuous variables of the model may assume negative values
            else:
                self.var_id2gpm[var_id] = gpm.addVar(lb=row['fbcLb'], ub=row['fbcUb'], vtype='C', name=var_id)

        # collect constraints with mapping to gp variables
        constrs = {}
        for var_id, row in self.m_dict['reactions'].iterrows():
            reactants = get_srefs(row['reactants'])
            products = get_srefs(row['products'])
            srefs = {constr_id: -coeff for constr_id, coeff in reactants.items()} | products
            for constr_id, coeff in srefs.items():
                if constr_id not in constrs:
                    constrs[constr_id] = self.var_id2gpm[var_id] * coeff
                else:
                    constrs[constr_id] += self.var_id2gpm[var_id] * coeff
                # implement splitting of transformed Gibbs energy of reaction variables
                if re.match(td_continuous_var_prefixes, var_id):
                    constrs[constr_id] += self.var_id2gpm[f'{var_id}_reverse'] * -coeff

        self._gp_add_add_linear_constraints(gpm, constrs)
        self._gp_add_optimization_objective(gpm)
        gpm.update()

        return gpm

    def _gp_create_model(self):
        """Create and configure a gurobipy model with data from metabolic model.

        - create empty gp.model:
        - add variables (using reaction data from metabolic model)
        - construct and implement linear constraints based on reaction string
        - configure TD binary variables and TD related inequality constraints
        - construct and implement linear objective

        :return: initialized gurobipy model
        :rtype: :class:`gurobipy.Model`
        """
        # create empty gurobipy model
        gpm = gp.Model(self.m_dict['modelAttrs'].id)

        # set model parameters
        gpm.setParam('OutputFlag', 0)
        # COBRApy default feasibility
        gpm.params.FeasibilityTol = 1e-7  # 1e-6 default
        gpm.params.OptimalityTol = 1e-7  # 1e-6 default
        gpm.params.IntFeasTol = 1e-7  # 1e-5 default

        # add variables to the gurobi model, with suppport of binary variables
        self.var_id2gpm = {}
        for var_id, row in self.m_dict['reactions'].iterrows():
            # configure TD forward/reverse use variables as binary
            if re.match(pf.V_FU_, var_id) or re.match(pf.V_RU_, var_id):
                self.var_id2gpm[var_id] = gpm.addVar(lb=0.0, ub=1.0, vtype='B', name=var_id)
            else:
                self.var_id2gpm[var_id] = gpm.addVar(lb=row['fbcLb'], ub=row['fbcUb'], vtype='C', name=var_id)

        # collect constraints with mapping to gp variables
        constrs = {}
        for var_id, row in self.m_dict['reactions'].iterrows():
            reactants = get_srefs(row['reactants'])
            products = get_srefs(row['products'])
            srefs = {constr_id: -coeff for constr_id, coeff in reactants.items()} | products
            for constr_id, coeff in srefs.items():
                if constr_id not in constrs:
                    constrs[constr_id] = self.var_id2gpm[var_id] * coeff
                else:
                    constrs[constr_id] += self.var_id2gpm[var_id] * coeff

        self._gp_add_add_linear_constraints(gpm, constrs)
        self._gp_add_optimization_objective(gpm)
        gpm.update()

        return gpm

    def _get_biomass_rid(self):
        """Extract biomass reaction id from model fbcObjectives.

        :return: biomass reaction id (without `R_`)
        :rtype: str
        """
        df_fbc_objs = self.m_dict['fbcObjectives']
        active_odata = df_fbc_objs[df_fbc_objs['active'] == bool(True)].iloc[0]
        sref_str = active_odata['fluxObjectives'].split(';')[0]
        return re.sub('^{pf.R_}', '', sbmlxdf.extract_params(sref_str)['reac'])

    # RETRIEVING DATA REQUIRED FOR RESULTS ANALYSIS
    @staticmethod
    def _get_reaction_str(rdata):
        """Generate the reactions string from a reaction object.

        e.g. for PFK_iso1:     'M_2pg_c => M_adp_c + M_fdp_c + M_h_c'
        e.g. for PFK_iso1_REV: 'M_adp_c + M_fdp_c + M_h_c => M_2pg_c'

        Exclude constraints (starting with `C_`)

        :param dict(str, dict) rdata: reaction data
        :return: reactions sting
        :rtype: str
        """
        lparts = []
        rparts = []
        for sid, stoic in get_srefs(rdata['reactants']).items():
            if re.match(pf.C_, sid) is None:
                # drop stoichiometric coefficients that are 1.0
                stoic_str = f'{round(stoic, 4)} ' if stoic != 1.0 else ''
                lparts.append(f'{stoic_str}{sid}')
        for sid, stoic in get_srefs(rdata['products']).items():
            if re.match(pf.C_, sid) is None:
                stoic_str = f'{round(stoic, 4)} ' if stoic != 1.0 else ''
                rparts.append(f'{stoic_str}{sid}')
        dirxn = ' -> ' if rdata['reversible'] else ' => '
        return ' + '.join(lparts) + dirxn + ' + '.join(rparts)

    def _expand_pseudo_metabolite(self, rid, reaction_str):
        """Expand pseudo metabolite in reaction string of iso reaction.

        Pseudo metabolites `pmet_` are used, when `_arm` reactions are introduced,
        e.g. in GECKO models. '_arm' reactions combine fluxes of iso-reactions.
        Iso reactions are connected via pseudo metabolites to the arm reaction

        E.g. reversible iML1515 reaction R_FBA in GECKO model
          FBA_arm, FBA_iso1, FBA_iso2
          FBA_arm_REV, FBA_iso1_REV, FBA_iso2_REV

        :param str rid: iso reaction idenitifer
        :param str reaction_str: reaction string
        :return: reactions string with metabolite
        :return: str
        """
        arm_rid = re.sub(r'_iso\d+', '_arm', rid)
        arm_rdata = self.m_dict['reactions'].loc[arm_rid]
        parts = re.split(' [-=]> ', self._get_reaction_str(arm_rdata))
        expand_pmet = parts[1] if re.search(r'\w*pmet_\w+', parts[0]) else parts[0]
        return re.sub(r'\w*pmet_\w+', expand_pmet, reaction_str)

    def _get_reaction_compartments(self, reaction_str):
        """From reaction string extract participating compartments

        e.g.
          reaction_str = 'M_2dmmq8_c + M_h2_c + 2.0 M_h_c => M_2dmmql8_c + 2.0 M_h_p'
          return: 'c_p'

        :param str reaction_str: species srefs string as per sbmlxdf reactants/products fields in reactions
        :return: sorted compartment ids joined by '-'
        :rtype: str
        """
        cids = set()
        srefs = parse_reaction_string(reaction_str)
        for idx, r_side in enumerate(['reactants', 'products']):
            for sref_str in sbmlxdf.record_generator(srefs[r_side]):
                params = sbmlxdf.extract_params(sref_str)
                sid = params['species']
                cids.add(self.m_dict['species'].loc[sid].compartment)
        return '-'.join(sorted(cids))

    def _get_reaction_data(self):
        """Extract reaction related data from reactions (reaction string, gpr)

        Supports ARM reactions (e.g. GECKO model) by expanding pseudo metabolites.
        Data can be used in augmenting flux related optimization results

        :return: reaction related data, indexed by reaction id
        :rtype: dict
        """
        gpid2label = self.m_dict['fbcGeneProducts']['label'].to_dict()
        rdatas = {}
        for rid, row in self.m_dict['reactions'].iterrows():
            if re.match(pf.V_, rid) is None and re.search('_arm', rid) is None:
                drxn = 'rev' if re.search('_REV$', rid) else 'fwd'
                fwd_rid = re.sub('_REV$', '', rid)
                net_rid = re.sub(r'_iso\d+', '', fwd_rid)
                groups = '; '.join(self.id2groups.get(rid, ''))
                reaction_str = self._get_reaction_str(row)
                if re.search(r'pmet_\w+', reaction_str) is not None:
                    reaction_str = self._expand_pseudo_metabolite(rid, reaction_str)
                # translate gene product ids to gene labels
                parts = []
                if type(row['fbcGeneProdAssoc']) is str:
                    assoc = re.sub('assoc=', '', row['fbcGeneProdAssoc'])
                    for item in re.findall(r'\b\w+\b', assoc):
                        if item in gpid2label:
                            parts.append(gpid2label[item])
                        elif item in {'and', 'or'}:
                            parts.append(item)
                gpr = ' '.join(parts)
                # mpmf coupling used in GECKO models for fitting kcats to proteomics (not required for RBA)
                mpmf_coupling = {}
                if gpr != '':
                    loci = [item.strip() for item in gpr.split(' and ')]
                    reactants = get_srefs(row['reactants'])
                    for locus in loci:
                        locus = valid_sbml_sid(locus)
                        # note: FBA/TFA models have not data in locus2uid
                        if locus in self.locus2uid:
                            ecm_coupling_constr_id = f'{pf.C_prot_}{self.locus2uid[locus]}'
                            if ecm_coupling_constr_id in reactants:
                                mpmf_coupling[locus] = reactants[ecm_coupling_constr_id]/self.protein_per_gdw

                exchange = False if type(row['products']) is str else True
                rp_cids = self._get_reaction_compartments(reaction_str)
                r_type = 'transport' if '-' in rp_cids else 'metabolic'

                rdatas[rid] = {'net_rid': net_rid, 'drxn': drxn, 'reaction_str': reaction_str,
                               'gpr': gpr, 'mpmf_coupling': mpmf_coupling,
                               'is_exchange': exchange, 'r_type': r_type, 'compartment': rp_cids,
                               'groups': groups}
        return rdatas

    def get_tx_metab_genes(self):
        """Get gene identifiers grouped by transport and metabolic reactions.

        Genes required in both transport and metabolic enzymes are designated transport related genes.

        Example: Extract sets of transport and metabolic related genes.

        .. code-block:: python

            tx_genes, metab_genes = eo.get_tx_metab_genes()

        :return: gene identifiers grouped by transport and metabolic enzymes
        :rtype: (set(str), set(str))
        """
        tx_genes = set()
        metab_genes = set()
        for data in self.rdata.values():
            if len(data['gpr']) > 0:
                gpr_del_or = re.sub(' or ', ',', data['gpr'])
                gpr_del_or_and = re.sub(' and ', ',', gpr_del_or)
                genes = [item.strip() for item in gpr_del_or_and.split(',')]
                if data['r_type'] == 'transport':
                    for gene in genes:
                        tx_genes.add(gene)
                elif data['r_type'] == 'metabolic':
                    for gene in genes:
                        metab_genes.add(gene)
        metab_genes = metab_genes.difference(tx_genes)
        return tx_genes, metab_genes

    @staticmethod
    def _extract_net_reaction_data(rdata):
        """Extract net reaction data, i.e. unsplit reactions

        Net reaction is a reaction with reaction string in forward direction
        Reversibility arrow is adjusted if there exists a corresponding reaction in reverse
        Gene Product Associations are combined using 'or' for iso-reactions

        :param dict(dict) rdata: reaction information
        :return: reaction data with isoreactions combined
        :rtype: dict(dict)
        """
        # determine 'net' reaction data (i.e. unsplit the reaction)
        rev_net_rids = {record['net_rid'] for rid, record in rdata.items() if record['drxn'] == 'rev'}
        net_rdata = {}
        for rid, record in rdata.items():
            if record['drxn'] == 'fwd':
                net_rid = record['net_rid']
                if net_rid not in net_rdata:
                    reaction_str = record['reaction_str']
                    if net_rid in rev_net_rids:
                        reaction_str = re.sub('=>', '->', reaction_str)
                    net_rdata[net_rid] = {'reaction_str': reaction_str, 'gpr': record['gpr'],
                                          'groups': record['groups']}
                else:
                    net_rdata[net_rid]['gpr'] += ' or ' + record['gpr']
        return net_rdata

    def get_rids_catalyzed(self):
        """Get reactions that are catalyzed with enzyme composition (gene labels).

        Data can be used to assess impact of gene deletions
        Data can be used to identify reactions catalyzed by a specific gene

        Note: does not support 'or' in alternative components of enzyme complex like (b1234 or b2346) and b2222
        Workaround: rearrange gpr: (b1234 and b2222) or (b2346 and b2222)

        :meta private:
        :return: enzymes and their compositions of catalyzed reactions
        :rtype: dict (key: rid/str, val: list of sets of gene labels)
        """
        gpid2label = self.m_dict['fbcGeneProducts']['label'].to_dict()
        rids_catalyzed = {}
        for rid, rdata in self.m_dict['reactions'].iterrows():
            if re.match(pf.V_, rid) is None:
                enzymes = []
                if type(rdata['fbcGeneProdAssoc']) is str:
                    assoc = re.sub('assoc=', '', rdata['fbcGeneProdAssoc'])
                    for enz_str in assoc.split('or'):
                        labels = set()
                        for gpid in re.findall(r'\b\w+\b', enz_str):
                            if gpid in gpid2label:
                                labels.add(gpid2label[gpid])
                        enzymes.append(labels)
                if len(enzymes) > 0:
                    rids_catalyzed[rid] = enzymes
        return rids_catalyzed

    # SUPPORT FUNCTIONS FOR GENE DELETION STUDY
    def get_rids_blocked_by(self, genes):
        """Identify reactions that get blocked when deleting selected genes.

        Example: Identify reactions that are blocked by the strain background.

        .. code-block:: python

            strain_background = {'b0063', 'b0062', 'b0061', 'b3904', 'b3903', 'b3902', 'b0344'}
            blocked_rids = eo.get_rids_blocked_by(strain_background)

        :param genes: gene or genes
        :type genes: str or list(str)
        :return: reaction ids affected by gene deletions
        :return: list(str)
        """
        if type(genes) is str:
            genes = {genes}
        elif type(genes) is list:
            genes = set(genes)
        assert type(genes) is set, 'genes must be of type str of a set/list of str'

        blocked_rids = []
        for rid, enzymes in self.rids_catalyzed.items():
            impacted = True
            for enzyme in enzymes:
                if len(genes.intersection(enzyme)) == 0:
                    impacted = False
                    break
            if impacted:
                blocked_rids.append(rid)
        return blocked_rids

    # OPTIMIZATION RELATED
    def get_variable_bounds(self, var_ids):
        """Retrieve upper/lower bounds of selected model variables.

        Bounds can be queried for a single variable or a list of variables.

        Example: Retrieve variable bounds for single variable, list of variables.

        .. code-block:: python

            bounds_1 = eo.get_variable_bounds('V_PC_total')
            bounds_2 = eo.get_variable_bounds(['POR5_iso1', 'POR5_iso1_REV'])

        :param var_ids: variable identifier or a list of variable identifiers
        :type var_ids: str or list(str)
        :return: variable ids with respective lower und upper bounds
        :rtype: dict(str, tuple)
        """
        if type(var_ids) is str:
            var_ids = [var_ids]

        variable_bounds = {}
        if self.is_gpm:
            self.gpm.update()
            #  support splitting of variables in two non-negative variables.
            for var_idx in var_ids:
                var_id = f'{pf.R_}{var_idx}' if f'{pf.R_}{var_idx}' in self.var_id2gpm else var_idx
                assert var_id in self.var_id2gpm, f'{var_id} not found'
                var = self.var_id2gpm[var_id]
                if var_id in self.gp_neg_var:
                    neg_var = self.var_id2gpm[self.gp_neg_var[var_id]]
                    ub = var.ub if neg_var.lb == 0.0 else -neg_var.lb
                    lb = -neg_var.ub if var.lb == 0.0 else var.lb
                else:
                    ub = var.ub
                    lb = var.lb
                variable_bounds[var_idx] = (lb, ub)
        else:
            for var_id in var_ids:
                assert var_id in self.model.reactions, f'{var_id} not found'
                variable_bounds[var_id] = self.model.reactions.get_by_id(var_id).bounds

        return variable_bounds

    def set_variable_bounds(self, variable_bounds):
        """Set upper/lower bounds of selected model variables.

        Bound values set to 'None' will not be modified.

        Example: Block iso-reaction POR5_iso1 in forward and reverse directions. Only upper bounds
        need to be changed to 0.0.

        .. code-block:: python

            orig_bounds = eo.set_variable_bounds({'POR5_iso1': (None, 0.0), 'POR5_iso1_REV': (None, 0.0)})

        :param dict(str, tuple) variable_bounds: variable identifiers with tuple of lower and upper bounds
        :return: original bounds
        :rtype: dict(str, tuple)
        """
        orig_bounds = self.get_variable_bounds(list(variable_bounds))
        if self.is_gpm:
            self.gpm.update()
            # support splitting of variables in two non-negative variables.
            for var_idx, (lb, ub) in variable_bounds.items():
                var_id = f'{pf.R_}{var_idx}' if f'{pf.R_}{var_idx}' in self.var_id2gpm else var_idx
                assert var_id in self.var_id2gpm, f'{var_id} not found'
                var = self.var_id2gpm[var_id]
                if var_id in self.gp_neg_var:
                    neg_var = self.var_id2gpm[self.gp_neg_var[var_id]]
                    if ub is not None:
                        if ub >= 0.0:
                            var.ub = ub
                            if neg_var.lb > 0.0:
                                neg_var.lb = 0.0
                        else:
                            neg_var.lb = -ub
                            if var.ub > 0.0:
                                var.ub = 0.0
                    if lb is not None:
                        if lb >= 0.0:
                            var.lb = lb
                            if neg_var.ub > 0.0:
                                neg_var.ub = 0.0
                        else:
                            neg_var.ub = -lb
                            if var.lb > 0.0:
                                var.lb = 0.0
                else:
                    if lb is not None:
                        var.lb = lb
                    if ub is not None:
                        var.ub = ub
                self.gpm.update()
        else:
            for var_id, (lb, ub) in variable_bounds.items():
                assert var_id in self.model.reactions, f'{var_id} not found'
                rxn = self.model.reactions.get_by_id(var_id)
                # orig_bounds[var_id] = rxn.bounds
                if (lb is not None) and (ub is not None):
                    rxn.bounds = (lb, ub)
                elif lb is not None:
                    rxn.lower_bound = lb
                elif ub is not None:
                    rxn.upper_bound = ub
        return orig_bounds

    def get_objective(self):
        """Retrieve model optimization objective.

        Example: Query optimization objective of the model.

        .. code-block:: python

            eo.get_objective()

        :return: variable coefficients and optimization direction
        :rtype: dict(str, float), str
        """
        if self.is_gpm:
            lin_expr = self.gpm.getObjective()
            objective = {}
            for idx in range(lin_expr.size()):
                var_id = lin_expr.getVar(idx).VarName
                objective[re.sub(f'^{pf.R_}', '', var_id)] = lin_expr.getCoeff(idx)
            direction = {gp.GRB.MINIMIZE: 'min', gp.GRB.MAXIMIZE: 'max'}[self.gpm.ModelSense]
        else:
            objective = {}
            sympy_expr = self.model.solver.objective.expression
            for var, coeff in sympy_expr.as_coefficients_dict().items():
                varname = var.name
                if '_reverse_' not in varname:
                    objective[varname] = coeff
            direction = self.model.objective.direction
        return objective, direction

    def set_objective(self, objective, direction='max'):
        """Set optimization objective.

        Example: Modify the optimization objective of a model.

        .. code-block:: python

            eo.set_objective({'BIOMASS_Ec_iML1515_WT_75p37M': 1.0}, 'max')

        :param dict(str, float) objective: variable identifiers with coefficients
        :param str direction: direction of optimization: 'min' or 'max' (default: 'max')
        """
        if self.is_gpm:
            sense = gp.GRB.MAXIMIZE if 'max' in direction.lower() else gp.GRB.MINIMIZE
            variables = []
            coeffs = []
            for var_id, coeff in objective.items():
                if var_id not in self.var_id2gpm:
                    var_id = f'{pf.R_}{var_id}'
                    assert var_id in self.var_id2gpm
                variables.append(self.var_id2gpm[var_id])
                coeffs.append(coeff)
            self.gpm.setObjective(gp.LinExpr(coeffs, variables), sense)
            self.gpm.update()
        else:
            self.model.objective = {self.model.reactions.get_by_id(var_id): coeff
                                    for var_id, coeff in objective.items()}
            self.model.objective_direction = direction

    def set_tfa_metab_concentrations(self, metab_concs):
        """Configure metabolite concentration ranges in mol/l for TD constraint models.

        Metabolomics data for a given condition can be used to further constrain models that
        contain thermodynamics constraints. Such models have optimization variables for
        species concentrations in log10. Minimum and maximum concentrations for selected species can be
        configured using this method. Concentration bounds will not be set for 'None' values.

        Example: Set minimum and maximum concentration values (mol/l) for selected metabolites.

        .. code-block:: python

            molar_concs = {'atp_c': (0.00178, 0.00712), 'adp_c': (5.8e-05, 0.00023), 'datp_c': (9.05e-05, 0.00036)}
            eo.set_tfa_metab_concentrations(molar_concs)

        :param dict(str, tuple) metab_concs: min and max concentrations for species in mol/l
        :return: original concentrations (min, max)
        :rtype: dict (str, tuple(float))
        """
        variable_bounds = {}
        for sid, (lb, ub) in metab_concs.items():
            sidx = re.sub(f'^{pf.M_}', '', sid)
            var_id = f'{pf.V_LC_}{sidx}'
            log10_lb = np.log10(lb) if lb else None
            log10_ub = np.log10(ub) if ub else None
            variable_bounds[var_id] = (log10_lb, log10_ub)

        orig_log10_bounds = self.set_variable_bounds(variable_bounds)

        orig_bounds = {}
        for var_id, (log10_lb, log10_ub) in orig_log10_bounds.items():
            sidx = re.sub(pf.V_LC_, '', var_id)
            lb = 10**log10_lb if log10_lb else None
            ub = 10**log10_ub if log10_ub else None
            orig_bounds[sidx] = (lb, ub)
        return orig_bounds

    def update_relaxation(self, fluxes, eps=0.5):
        """TFA slack model relaxation for a specified flux distribution.

        In case optimizations of a model with added thermodynamics constraints leads
        to infeasable solutions, model relaxation is required. Model relaxation is performed on a
        TFA model with slack variables configured for ∆rG'˚ (standard transformed Gibbs energy of reaction)
        variables. Use TfaModel.export_slack_model() to generate such a model.
        This method is used to determine adjusted ∆rG'˚ variable bounds that make the TD model feasible
        and to configure the updated bounds to support optimization of further conditions.

        Example 1: Perform TFA model relaxation using COBRApy interface. With the additional constraint
        of achieving at least 95% of the growth rate of the model without TD constraints.

        .. code-block:: python

            import cobra
            from f2xba import FbaOptimization

            tfa_slack_model = cobra.io.read_sbml_model('iML1515_TFA_slack.xml')
            fo = FbaOptimization('iML1515_TFA_slack.xml', tfa_slack_model)

            tfa_slack_model.reactions.get_by_id('BIOMASS_Ec_iML1515_core_75p37M').lower_bound = 0.95 * wt_gr
            tfa_slack_model.medium = medium

            solution = tfa_slack_model.optimize()
            drg0_relaxations = fo.update_relaxation(solution.fluxes)

        Example 2: Perform TFA model relaxation using cobrapy interface.

        .. code-block:: python

            from f2xba import FbaOptimization

            fo = FbaOptimization('iML1515_TFA_slack.xml')

            orig_bounds = fo.set_variable_bounds({'BIOMASS_Ec_iML1515_core_75p37M':(0.95 * wt_gr, None)})
            fo.medium = medium

            solution = fo.optimize()
            fo.set_variable_bounds(orig_bounds)
            drg0_relaxations = fo.update_relaxation(solution.fluxes)

        :param dict fluxes: fluxes of a TFA slack model optimization solution
        :param float eps: additional margin in kJ/mol (default: 0.5)
        :return: ∆rG'˚ variables with updated bounds.
        :rtype: dict(str, dict(str,float)
        """
        # collect DRG0 variables that need relaxation
        slack_vars = {}
        for var_id, val in fluxes.items():
            if re.match(pf.V_NS_, var_id) and val > 1e-7:
                slack_vars[re.sub(f'^{pf.V_NS_}', pf.V_DRG0_, var_id)] = -val
            elif re.match(pf.V_PS_, var_id) and val > 1e-7:
                slack_vars[re.sub(f'^{pf.V_PS_}', pf.V_DRG0_, var_id)] = val

        # relax DRG0 variables by adjusting respective bound
        drg0_relaxations = {}
        update_bounds = {}
        for drg0_id, (lb, ub) in self.get_variable_bounds(list(slack_vars)).items():
            slack = slack_vars[drg0_id]
            if slack < 0.0:
                new_lb = lb + slack - eps
                update_bounds[drg0_id] = (new_lb, None)
                drg0_relaxations[drg0_id] = {'fbc_lower_bound': new_lb}
            else:
                new_ub = ub + slack + eps
                update_bounds[drg0_id] = (None, new_ub)
                drg0_relaxations[drg0_id] = {'fbc_upper_bound': new_ub}

        self.set_variable_bounds(update_bounds)
        return drg0_relaxations

    def modify_stoic(self, constr_id, var_id, new_val):
        """Modify a stoichiometric coefficient using gurobipy interface.

        Stoichiometric coefficients in a COBRApy model can be modified using the COBRApy method
        cobra.core.reaction.add_metabolites().
        This method supports the modification of a stoichiometric coefficient
        when using the gurobipy interface. As model contexts are not supported, the coefficient
        should be reset to the original value after optimization.

        Example: Remove the cofactor requirement of hemeA in anaerobic conditions.

        .. code-block:: python

            if 'EX_o2_e' not in medium:
                orig_stoic = eo.modify_stoic('M_hemeA_c', 'R_CofactorSynth', 0.0)

            eo.optimize()

            if 'EX_o2_e' not in medium:
                eo.modify_stoic('M_hemeA_c', 'R_CofactorSynth', orig_stoic)

        :param str constr_id: constraint identifier, e.g. species id
        :param str var_id: variable identifier, e.g. reaction id
        :param float new_val: stoichiometric coefficient (negative for consumation)
        :return: original stoichiometric coefficient
        :rtype: float
        """
        if self.is_gpm:
            constr = self.gpm.getConstrByName(constr_id)
            var = self.gpm.getVarByName(var_id)
            if constr is None:
                print(f'Constraint with identifier "{constr_id}" not found!')
                return None
            if var is None:
                print(f'Variable with identifier "{var_id}" not found!')
                return None
            old_val = self.gpm.getCoeff(constr, var)
            self.gpm.chgCoeff(constr, var, new_val)
            if var_id in self.gp_neg_var:
                neg_var = self.gpm.getVarByName(self.gp_neg_var[var_id])
                self.gpm.chgCoeff(constr, neg_var, -new_val)

            return old_val
        else:
            print('Method only implemented for gurobipy interface, for COBRApy, use rxn.add_metabolites()')
            return None

    def set_medium(self, medium):
        """Configure medium in gurobipy model from dictionary with uptake fluxes.

        It is recommended to rather assign the medium as in COBRApy (use medium property)

        e.g. medium = {'EX_glc__D_e': 10.0, 'EX_o2_e': 20.0, 'EX_ca2_e': 1000.0, ...}

        For gurobipy implementation, support splitting of variables in two non-negative variables.

        :meta private:
        :param dict(str, float) medium: nutrients in medium, i.e. exchange reactions with max uptake rate
        """
        if self.is_gpm:
            update_bounds = {}
            for ex_rid in self.uptake_rids:
                ex_ridx = re.sub(f'^{pf.R_}', '', ex_rid)
                if ex_ridx in medium:
                    update_bounds[ex_rid] = (-medium[ex_ridx], None)
                else:
                    update_bounds[ex_rid] = (0.0, None)
            self.set_variable_bounds(update_bounds)
        else:
            self.model.medium = medium

    def optimize(self, alt_model=None):
        """Optimize gurobipy model and return COBRApy compliant solution.

        Example: Optimize the model.

        .. code-block:: python

            eo.optimize()

        Thermodynamics constrained models are supported inline with pyTFA package.

        :param alt_model: gurobipy model (default: None)
        :type alt_model: :class:`gurobipy.Model`
        :return: optimization solution
        :rtype: :class:`Solution`
        """
        model = self.gpm if alt_model is None else alt_model

        model.optimize()
        solution = self._gp_get_solution(model)

        return solution

    def _gp_combine_neg_vars(self, data_raw):
        """Combine negative variable values from gurobipy optimization.

        :param dict(str, float) data_raw: optimization variables with values.
        :return:
        """
        data = {}
        neg_var_ids = set(self.gp_neg_var.values())
        for var_id, val in data_raw.items():
            var_idx = re.sub(f'^{pf.R_}', '', var_id)
            if var_id not in neg_var_ids:
                if var_id in self.gp_neg_var:
                    data[var_idx] = data_raw[var_id] - data_raw[self.gp_neg_var[var_id]]
                else:
                    data[var_idx] = data_raw[var_id]
        return data

    def _gp_get_solution(self, alt_model=None):
        """Retrieve optimization solution for gp model.

        Solution as per COBRApy solution
        Reaction and Metabolites ids without `R_`, `M_` prefix
        - status
        - objective value
        - fluxes
        - reduced costs
        - shadow prices

        :param alt_model: (optional) alternative model (e.g. pFBA model)
        :type alt_model: gurobipy.Model, if provided
        :return: optimization solution
        :rtype: class:`Solution`
        """
        model = self.gpm if alt_model is None else alt_model

        results_dict = {'status': status2text[model.status].lower(), 'tmp': pd.Series()}
        if hasattr(model, 'objval'):
            results_dict['objective_value'] = model.objval
        if hasattr(model.getVars()[0], 'x'):
            data_raw = {var.VarName: var.X for var in model.getVars()}
            results_dict['fluxes'] = pd.Series(self._gp_combine_neg_vars(data_raw), name='fluxes')
        if hasattr(model.getVars()[0], 'rc'):
            data_raw = {var.VarName: var.RC for var in model.getVars()}
            results_dict['reduced_costs'] = pd.Series(self._gp_combine_neg_vars(data_raw), name='reduced_costs')
        if hasattr(model.getConstrs()[0], 'pi'):
            data = {re.sub(f'^{pf.M_}', '', constr.ConstrName): constr.Pi
                    for constr in model.getConstrs()}
            results_dict['shadow_prices'] = pd.Series(data, name='shadow_prices')
        del results_dict['tmp']
        return Solution(**results_dict)

    def pfba(self, fraction_of_optimum=1.0):
        """Perform parsimonious FBA optimization (pFBA) on a FBA model using the gurobipy interface.

        Example: Execute a pFBA optimization using the gurobipy interface.

        .. code-block:: python

            solution = eo.pfba()

        :param float fraction_of_optimum: factor to scale FBA objective (default: 1.0)
        :return: optimization solution
        :return: :class:`Solution`
        """
        if self.is_gpm is None:
            print('Method implemented for gurobipy interface only.')
            return None

        # determine fba growth rate
        fba_solution = self.optimize()
        fba_gr = fba_solution.objective_value

        pfba_solution = fba_solution.objective_value
        if np.isfinite(fba_gr):

            # create an pFBA model based on FBA model and make reactions irreversible
            self.gpm.copy()
            pfba_gpm = self.gpm.copy()

            for var in pfba_gpm.getVars():
                if var.lb < 0:
                    fwd_col = pfba_gpm.getCol(var)
                    rev_coeffs = [-fwd_col.getCoeff(idx) for idx in range(fwd_col.size())]
                    constrs = [fwd_col.getConstr(idx) for idx in range(fwd_col.size())]
                    rev_col = gp.Column(coeffs=rev_coeffs, constrs=constrs)
                    pfba_gpm.addVar(lb=0.0, ub=-var.lb, vtype=var.VType, name=f'{var.VarName}_REV', column=rev_col)
                    var.lb = 0.0

            # create temporary constraint for FBA objective
            fba_objective = pfba_gpm.getObjective()
            pfba_gpm.addLConstr(fba_objective, '>', fba_gr * fraction_of_optimum, 'FBA objective')

            # New Objective: minimize sum of all non-negative reactions
            pfba_gpm.update()
            vars_gpm = [var for var in pfba_gpm.getVars()]
            pfba_objective = gp.LinExpr(np.ones(len(vars_gpm)), vars_gpm)
            pfba_gpm.setObjective(pfba_objective, gp.GRB.MINIMIZE)

            # optimize and collect results (without error handling)
            pfba_gpm.optimize()
            pfba_solution = self._gp_get_solution(pfba_gpm)
            pfba_solution.fluxes = extract_net_fluxes(pfba_solution.fluxes)
            pfba_gpm.close()

        return pfba_solution

    def fva(self, rids=None, fraction_of_optimum=1.0, net=False, cutoff=None):
        """Perform a Flux Variability Analysis (FVA) on a FBA model using the gurobipy interface.

        A list of reaction ids can be provided to limit the scope of the FVA analysis.
        With the optional parameter 'net' set to True, fluxes of isoreactions and forward/reverse
        are combined. With optional parameter 'cutoff', absolute flux values below the cutoff, e.g. 1e-8,
        are considered to be zero.

        Example: Execute FVA analyis for a set of reactions using the gurobipy interface.

        .. code-block:: python

            eo.fva(['CS', 'ME1', 'POR5_iso1', 'POR5_iso1_REV'])

        :param list(str) rids: reaction identifiers (without leading `R_`) (default: None)
        :param float fraction_of_optimum: scaling of wild type objective value (default: 1.0)
        :param bool net: (optional) select if fluxes of isoreactions and rev/fwd should be aggregated
        :param float cutoff: (optional) cutoff value for fluxes to be considered to be zero, e.g. 1e-8
        :return: table with minimum and maximum reaction fluxes
        :rtype: pandas.DataFrame
        """
        if self.is_gpm is None:
            print('Method implemented for gurobipy interface only.')
            return None

        ridx2rid = {re.sub(f'^{pf.R_}', '', rid): rid
                    for rid in self.m_dict['reactions'].index if re.match(pf.R_, rid)}
        selected_rids = list(ridx2rid.values()) if rids is None else [ridx2rid[ridx] for ridx in rids]

        wt_gr = self.optimize().objective_value
        results = {}
        if np.isfinite(wt_gr):

            # create a temporary FVA model based on FBA model
            self.gpm.update()
            fva_gpm = self.gpm.copy()

            # add wt objective as constraint
            wt_objective = fva_gpm.getObjective()
            fva_gpm.addLConstr(wt_objective, '>', wt_gr * fraction_of_optimum, 'wild type objective')
            fva_gpm.update()

            for rid in selected_rids:
                var = fva_gpm.getVarByName(rid)

                fva_gpm.setObjective(var, gp.GRB.MINIMIZE)
                min_flux = self.optimize(fva_gpm).objective_value

                fva_gpm.setObjective(var, gp.GRB.MAXIMIZE)
                max_flux = self.optimize(fva_gpm).objective_value

                if cutoff:
                    if abs(max_flux) < cutoff:
                        max_flux = 0.0
                    if abs(min_flux) < cutoff:
                        min_flux = 0.0

                results[re.sub(f'^{pf.R_}', '', rid)] = [min_flux, max_flux]
            fva_gpm.close()

        net_rids = set()
        if not net:
            df_fva_results = pd.DataFrame(results.values(), list(results.keys()), columns=['minimum', 'maximum'])
        else:
            net_flux_ranges = {}
            for rid, (flux_min, flux_max) in results.items():
                if re.match(pf.V_, rid) is None and re.search(r'_arm(_REV)?$', rid) is None:
                    fwd_rid = re.sub('_REV$', '', rid)
                    net_rid = re.sub(r'_iso\d*', '', fwd_rid)
                    net_rids.add(net_rid)
                    if flux_max > 0.0:
                        if re.search('_REV$', rid):
                            net_flux_min = -flux_max
                            net_flux_max = -flux_min
                        else:
                            net_flux_min = flux_min
                            net_flux_max = flux_max
                        if net_rid in net_flux_ranges:
                            net_flux_min = min(net_flux_ranges[net_rid][0], net_flux_min)
                            net_flux_max = max(net_flux_ranges[net_rid][1], net_flux_max)
                        net_flux_ranges[net_rid] = [net_flux_min, net_flux_max]
            # add ranges for blocked net reactions
            for net_rid in net_rids:
                if net_rid not in net_flux_ranges:
                    net_flux_ranges[net_rid] = [0.0, 0.0]

            df_fva_results = pd.DataFrame(net_flux_ranges.values(), list(net_flux_ranges.keys()),
                                          columns=['minimum', 'maximum'])

        return df_fva_results

    def solve(self, **rba_solve_params):
        """RBA solver, overwritten by RbaOptimization

        :meta private:
        """
        return pd.DataFrame()

    def get_active_genes(self, fluxes):
        """Determine active genes based on a reaction flux distribution.

        All genes required for isoenzymes of a reactions carrying flux are collected.

        :meta private:
        :param dict(str, float) fluxes: reaction fluxes in mmol/gDWh
        :return: set of genes potentially required for given reaction flux destribution
        :rtype: set
        """
        # map reaction ids without 'R_' prefix, as per COBRApy to reaction ids used in the model
        ridx2rid = {re.sub(f'^{pf.R_}', '', rid): rid for rid in self.rids_catalyzed}

        gene_list = set()
        for ridx, flux in fluxes.items():
            if abs(flux) > 0.0:
                if ridx in ridx2rid:
                    for enz_comp in self.rids_catalyzed[ridx2rid[ridx]]:
                        for gene in enz_comp:
                            gene_list.add(gene)
        print(f'{len(gene_list)} genes potentially active given flux distribution.')
        return gene_list

    def gene_knock_outs(self, genes):
        """Block reactions that are catalyzed by specified gene/genes.

        Example: Block reactions that depend on the specified gene and perform a model optimization.

        .. code-block:: python

            orig_bounds = eo.gene_knock_outs('b0025')
            solution = eo.optimize()
            eo.set_variable_bounds(orig_bounds)
            print(f'mutant gr: {solution.objective_value:.3f} h-1')

        :param genes: gene identifiers
        :type genes: str or list(str)
        :return: original reaction bounds (lb, ub)
        :rtype: dict(str, tuple)
        """
        if type(genes) is str:
            genes = [genes]

        if self.is_gpm:
            update_bounds = {var_id: (0.0, 0.0) for var_id in self.get_rids_blocked_by(genes)}
            orig_rid_bounds = self.set_variable_bounds(update_bounds)
        else:
            orig_rid_bounds = {}
            for gene in genes:
                assert gene in self.model.genes, f'{gene} not found'
                for react in self.model.genes.get_by_id(gene).reactions:
                    orig_rid_bounds[react.id] = (react.lower_bound, react.upper_bound)
            for gene in genes:
                self.model.genes.get_by_id(gene).knock_out()

        return orig_rid_bounds

    def moma(self, wt_fluxes, linear=False):
        """Perform a MOMA based optimization using the gurobipy interface.

        Supported for FBA and ECM models.

        Ref: Segre et al. 2002, Analysis of optimality in natural and perturbed
        metabolic networks.

        MOMA (`linear` = False):

            minimize: ∑ (v - wt)^2

            s.t.: S * v = 0; v_lb ≤ v ≤ v_ub

            v: disturbed flux distribution; wt: wild type flux distribution.

            This requires adding transformed variables `vt` with adjusted bounds, adding new constraints `vt = v - wt`
            and minimizing the objective function `∑ vt^2`.

        Linear MOMA (`linear` = True):

            minimize: ∑ abs(v - wt)

            s.t.: S * v = 0; v_lb ≤ v ≤ v_ub

            introducing non-negative transformed flux variables and related constraints and minimizing `∑ (vtp + vtn)`

            `vtp`: [0 ... (v_ub - wt)] with constraints `vtp ≥ v - wt`

            `vtn`: [0 ... (wt - v_lb)] with constraints `vtn ≥ wt - v`

        Example: Perform MOMA analysis for a disturbed network, caused by gene deletion.

        .. code-block:: python

            fo.medium = medium
            wt_solution = fo.pfba()

            orig_bounds = fo.gene_knock_outs('b0003')
            moma_solution = fo.moma(wt_solution.fluxes)
            fo.set_variable_bounds(orig_bounds)
            print(f'{moma_solution.fluxes['BIOMASS_Ec_iML1515_core_75p37M']:.3f} h-1')

        :param dict(str, float) wt_fluxes: wild type flux distribution, e.g. pFBA fluxes
        :param bool linear: using quadratic (Euclidean distance) or linear formulation (Manhattan) (default: False)
        :return: MOMA determined solution
        :rtype: :class:`Solution`
        """
        if self.is_gpm is None:
            print('Method implemented for gurobipy interface only.')
            return pd.DataFrame()

        self.gpm.update()
        moma_gpm = self.gpm.copy()

        tflux_vars = []
        for ridx, wt_flux in wt_fluxes.items():
            flux_var = moma_gpm.getVarByName(f'{pf.R_}{ridx}')
            if flux_var is None:
                flux_var = moma_gpm.getVarByName(ridx)

            if not linear:
                tflux_lb = flux_var.LB - wt_flux
                tflux_ub = flux_var.UB - wt_flux
                tflux_var = moma_gpm.addVar(lb=tflux_lb, ub=tflux_ub, vtype=gp.GRB.CONTINUOUS,
                                            name=f'V_MOMA_transformed_flux_{ridx}')
                tflux_vars.append(tflux_var)
                moma_gpm.addLConstr(flux_var - tflux_var, sense=gp.GRB.EQUAL, rhs=wt_flux)
            else:
                tfluxp_ub = flux_var.UB - wt_flux
                # if wild type flux already at the upper boundary we do not require fluxp variable/constraint
                if tfluxp_ub > 0:
                    tfluxp_var = moma_gpm.addVar(lb=0, ub=tfluxp_ub, vtype=gp.GRB.CONTINUOUS,
                                                 name=f'V_MOMA_transformed_fluxp_{ridx}')
                    tflux_vars.append(tfluxp_var)
                    moma_gpm.addLConstr(flux_var - tfluxp_var, sense=gp.GRB.LESS_EQUAL, rhs=wt_flux)
                tfluxn_ub = wt_flux - flux_var.LB
                if tfluxn_ub > 0:
                    tfluxn_var = moma_gpm.addVar(lb=0, ub=tfluxn_ub, vtype=gp.GRB.CONTINUOUS,
                                                 name=f'V_MOMA_transformed_fluxn_{ridx}')
                    tflux_vars.append(tfluxn_var)
                    moma_gpm.addLConstr(flux_var + tfluxn_var, sense=gp.GRB.GREATER_EQUAL, rhs=wt_flux)

        if not linear:
            moma_gpm.setObjective(gp.quicksum([v * v for v in tflux_vars]), gp.GRB.MINIMIZE)
        else:
            moma_gpm.setObjective(gp.quicksum(tflux_vars), gp.GRB.MINIMIZE)

        moma_gpm.optimize()
        moma_solution = self._gp_get_solution(moma_gpm)
        moma_gpm.close()

        return moma_solution

    def room(self, wt_fluxes, linear=False, delta=0.03, epsilon=1e-3, time_limit=30.0):
        """Perform a ROOM based optimization using the gurobipy interface.

        Ref: Shlomi et al., 2005, Regulatory on off minimization of metabolic flux
        changes after genetic perturbations


        ROOM (`linear` = False):

            minimize: ∑ y^i

            s.t. S * v = 0; v_lb ≤ v ≤ v_ub

            v - y * (v_ub - wt_u) <= wt_u; wt_u = wt + delta * abs(wt) + epsilon

            v - y * (v_lb - wt_l) <= wt_l; wt_l = wt - delta * abs(wt) - epsilon

            v: disturbed flux distribution; wt: wild type flux distribution.

            y_i ∈ {0,1}

        Linear ROOM (`linear` = True):

            0 <= y_i <= 1

        Example: Perform ROOM analysis for a disturbed network, caused by gene deletion.

        .. code-block:: python

            fo.medium = medium
            wt_solution = fo.pfba()

            orig_bounds = fo.gene_knock_outs('b0003')
            room_solution = fo.room(wt_solution.fluxes)
            fo.set_variable_bounds(orig_bounds)
            print(f'{room_solution.fluxes['BIOMASS_Ec_iML1515_core_75p37M']:.3f} h-1')

        :param dict(str, float) wt_fluxes: wild type flux distribution, e.g. pFBA fluxes
        :param bool linear: using MILP (False) or relaxed LP (True) formulation (default: False)
        :param float delta: relative tolerance range (default: 0.03)
        :param float epsilon: absolute tolerance range (default: 1e-3)
        :param float time_limit: time limit in seconds (default: 30.0)
        :return: ROOM determined solution
        :rtype: :class:`Solution`
        """
        if self.is_gpm is None:
            print('Method implemented for gurobipy interface only.')
            return None

        self.gpm.update()
        room_gpm = self.gpm.copy()
        if not linear:
            room_gpm.params.timelimit = time_limit

        flux_ctrl_vars = []
        for ridx, wt_flux in wt_fluxes.items():
            flux_var = room_gpm.getVarByName(f'{pf.R_}{ridx}')
            if flux_var is None:
                flux_var = room_gpm.getVarByName(ridx)
            flux_lb = flux_var.LB
            flux_ub = flux_var.UB

            if not linear:
                wl_flux = wt_flux - abs(wt_flux) * delta - epsilon
                wu_flux = wt_flux + abs(wt_flux) * delta + epsilon
            else:
                wl_flux = wt_flux
                wu_flux = wt_flux
            if wl_flux > flux_lb:
                if not linear:
                    flux_lower_ctrl_var = room_gpm.addVar(lb=0, ub=1, vtype=gp.GRB.BINARY,
                                                          name=f'V_ROOM_wl_control_{ridx}')
                else:
                    flux_lower_ctrl_var = room_gpm.addVar(lb=0.0, ub=1.0, vtype=gp.GRB.CONTINUOUS,
                                                          name=f'V_ROOM_wl_control_{ridx}')
                flux_ctrl_vars.append(flux_lower_ctrl_var)
                room_gpm.addLConstr(flux_var - flux_lower_ctrl_var * (flux_lb - wl_flux),
                                    sense=gp.GRB.GREATER_EQUAL, rhs=wl_flux)
            if wu_flux < flux_ub:
                if not linear:
                    flux_upper_ctrl_var = room_gpm.addVar(lb=0, ub=1, vtype=gp.GRB.BINARY,
                                                          name=f'V_ROOM_wu_control_{ridx}')
                else:
                    flux_upper_ctrl_var = room_gpm.addVar(lb=0.0, ub=1.0, vtype=gp.GRB.CONTINUOUS,
                                                          name=f'V_ROOM_wu_control_{ridx}')
                flux_ctrl_vars.append(flux_upper_ctrl_var)
                room_gpm.addLConstr(flux_var - flux_upper_ctrl_var * (flux_ub - wu_flux),
                                    sense=gp.GRB.LESS_EQUAL, rhs=wu_flux)
        room_gpm.setObjective(gp.quicksum(flux_ctrl_vars), gp.GRB.MINIMIZE)

        room_gpm.optimize()
        room_solution = self._gp_get_solution(room_gpm)
        room_gpm.close()

        return room_solution

    def __del__(self):
        if self.is_gpm and self.gpm is not None:
            self.gpm.close()



"""Implementation of UniprotData class.

Peter Schubert, CCB, HHU Duesseldorf, January 2023
"""

import os
import re
import pandas as pd
import gzip
import urllib.parse
import urllib.request
import requests
from requests.adapters import HTTPAdapter

from .uniprot_protein import UniprotProtein

uniprot_rest_url = 'https://rest.uniprot.org'
uniprot_stream_path = '/uniprotkb/stream'
uniprot_search_path = '/uniprotkb/search'


def get_next_link(headers):
    """Retrieve from requests header the URL to download next segment of data.

    Implement pagination to retrieve larger number of results
    copied from https://www.uniprot.org/help/api_queries

    :meta private:
    :param headers: requests.structures.CaseInsensitiveDict
    :return:
    """
    if "Link" in headers:
        match = re.match(r'<(.+)>; rel="next"', headers["Link"])
        if match:
            return match.group(1)
    return None


def get_batch(session, batch_url):
    """Yield next download segment for processing.

    Terminate, once no more linked segments are in the download stream.

    Implement pagination to retrieve larger number of results

    copied from https://www.uniprot.org/help/api_queries

    :meta private:
    :param session: Requests session used to retrieve data
    :type session: :class:`requests.Session`
    :param str batch_url: URL for new requests object
    :return: download segment and total results so far
    :rtype: :class:`requests.Response`, int
    """
    while batch_url:
        response = session.get(batch_url)
        response.raise_for_status()
        total = response.headers["x-total-results"]
        yield response, total
        batch_url = get_next_link(response.headers)


class UniprotData:
    """Access to data of UniProt online resource.

    Enzyme constraint and resource balance constraint models require access to protein information
    during model construction. Protein related information can be retrieved from the UniProt online
    database (uniprot.org) using the taxonomic id (parameter `organism_id`) as reference.

    In case UniProt holds no data for the modelled organism, it is required to
    compile protein related data from other sources, with data fields and format as in a UniProt download,
    and store this file named `uniprot_organism_<organism_id>.tsv` under `organism_dir`.

    Use configuration data in the XBA configuration file, sheet `general`, to configure `organism_id`, with
    the taxonomic id, and `organism_dir`, were downloaded data is stored locally.
    Delete the locally stored UniProt data to enforce a retrieval from the online database.

    Example: Access UniProt data for E. coli K-12 MG1655 strain (taxonomic id: 83333).

    .. code-block:: python

        from f2xba.uniprot.uniprot_data import UniprotData

        uniprot_data = UniprotData(83333, 'data_ref')

        gene = 'b0928'
        uid = uniprot_data.locus2uid[gene]
        uniprot_data.proteins[uid].__dict__

    :param int or str organism_id: taxonomic identifier of modelled organism
    :param str organism_dir: directory where UniProt data is stored
    """

    def __init__(self, organism_id, organism_dir):
        """Initialize

        Download UniProt information for specific organism, if download
        not found in uniprot_dir.

        Processed UniProt export to extract protein information

        :param int or str organism_id: taxonomic identifier of modelled organism
        :param str organism_dir: directory where UniProt exports are stored
        """
        self.organism_id = organism_id
        self.fname = os.path.join(organism_dir, f'uniprot_organism_{organism_id}.tsv')

        if not os.path.exists(self.fname):
            self.download_data()
        else:
            print(f'extracting UniProt protein data from {self.fname}')

        df_uniprot = pd.read_csv(self.fname, sep='\t', index_col=0)

        self.proteins = {}
        """Protein related information extracted from UniProt."""

        for uid, row in df_uniprot.iterrows():
            self.proteins[uid] = UniprotProtein(row)

        self.locus2uid = {}
        """Map gene locus to UniProt identifyer of related protein."""

        self.update_locus2uid()

    def update_locus2uid(self):
        self.locus2uid = {}
        for uid, p in self.proteins.items():
            for locus in p.loci:
                self.locus2uid[locus] = uid
        return self.locus2uid

    def download_data(self):
        """Download required protein data from UniProt database

        Data is stored in 'self.uniprot_dir' in .tsv format.
        Based on  https://www.uniprot.org/help/api_queries,
        Query fields as per https://www.uniprot.org/help/return_fields
        using pagination and compression.

        :meta private:
        """
        # query = [f'(organism_id:{self.organism_id})', '(reviewed:true)']
        query = [f'(organism_id:{self.organism_id})']
        fields = ['accession', 'gene_primary', 'gene_synonym', 'gene_oln', 'organism_id',
                  'ec', 'protein_name', 'cc_subunit', 'cc_subcellular_location', 'cc_cofactor',
                  'length', 'mass', 'sequence', 'ft_signal', 'cc_catalytic_activity', 'kinetics',
                  'go_p', 'go_c', 'go_f', 'protein_families',
                  'xref_biocyc', 'xref_refseq', 'xref_kegg', 'date_modified', 'reviewed']
        payload = {'compressed': 'true',
                   'fields': ','.join(fields),
                   'format': 'tsv',
                   'query': ' AND '.join(query),
                   'size': 500,
                   }
        # extraction code based on UniProt example from https://www.uniprot.org/help/api_queries
        retries = requests.adapters.Retry(total=5, backoff_factor=0.25,
                                          status_forcelist=[500, 502, 503, 504])
        session = requests.Session()
        session.mount("https://", requests.adapters.HTTPAdapter(max_retries=retries))

        url = uniprot_rest_url + uniprot_search_path + '?' + urllib.parse.urlencode(payload)
        progress = 0
        with open(self.fname, 'w') as f:
            for batch, total in get_batch(session, url):
                records = (gzip.decompress(batch.content)).decode().splitlines()
                # drop header lines in subsequent segments
                idx = 0 if progress == 0 else 1
                for line in records[idx:]:
                    f.write(line + '\n')
                progress += len(records) - 1
                # print(f'{progress} / {total}')
            print(f'UniProt protein data downloaded for organism {self.organism_id} to: {self.fname}')

    def modify_attributes(self, df_modify_attrs):
        """Modify locus information for selected UniProt ids.

        UniProt loci might be missing in UniProt export,
            e.g. 'P0A6D5' entry has missing locus (as per July 2023)

        :meta private:
        :param pandas.DataFrame df_modify_attrs: data to be modified on proteins
        """
        for uid, row in df_modify_attrs.iterrows():
            if uid not in self.proteins:
                print(f'{uid} not found in UniProt data export')
            else:
                self.proteins[uid].modify_attribute(row['attribute'], row['value'])

        self.update_locus2uid()



"""Implementation of utility for single gene knockout analysis.

Peter Schubert, HHU Duesseldorf, November 2025
"""

import numpy as np
import pandas as pd

pred_outcome = {True: {True: 'tp', False: 'fn'},
                False: {True: 'fp', False: 'tn'}}


def confusion_matrix(act_classification, pred_classification):
    """Create a 2D confusion matrix based on actual and predicted classifications.

    Statistics, set of items and confusion matrix are returned in a dictionary.

    Example: Perform single gene deletion simulation (using gurobipy interface) and plot confusion matrix.
    `keio_ess` and `keio_red` hold lists of genes that are considered essential/redundant for selected condition.

    .. code-block:: python

        eo = EcmOptimization('iML1515_GECKO.xml')
        eo.medium = {rid: 1000.0 for rid in lb_medium}
        df_sgko = eo.single_gene_deletion()

        act_classification = {gene: False for gene in keio_red}
        act_classification.update({gene: True for gene in keio_ess})
        pred_classification = (df_sgko['fitness'] < 0.05).to_dict()
        pred = confusion_matrix(act_classification, pred_classification)

        print('recall:', pred['recall'])
        pred['cm']

    :param dict(str, bool) act_classification: actual classifications
    :param dict(str, bool) pred_classification: predicted classifications
    :return: prediction results
    :rtype: dict
    """
    pred_results = {'tp': set(), 'fn': set(), 'fp': set(), 'tn': set(),
                    'recall': 0.0, 'precision': 0.0, 'specificity': 0.0, 'accuracy': 0.0, 'mcc': 0.0,
                    'cm': pd.DataFrame()}
    for gene, pred_class in pred_classification.items():
        if gene in act_classification:
            prediction = pred_outcome[act_classification[gene]][pred_class]
            pred_results[prediction].add(gene)

    tp = len(pred_results['tp'])
    fn = len(pred_results['fn'])
    fp = len(pred_results['fp'])
    tn = len(pred_results['tn'])
    pred_results['recall'] = tp / (tp + fn) if (tp + fn) > 0 else np.nan
    pred_results['precision'] = tp / (tp + fp) if (tp + fp) > 0 else np.nan
    pred_results['specificity'] = tn / (fp + tn) if (fp + tn) > 0 else np.nan
    pred_results['accuracy'] = (tp + tn) / (tp + fn + fp + tn) if (tp + fn + fp + tn) > 0 else np.nan
    pred_results['mcc'] = ((tp * tn - fp * fn) / np.sqrt((tp + fp) * (fn + tn) * (tp + fn) * (fp + tn))
                           if (tp + fp) * (fn + tn) * (tp + fn) * (fp + tn) > 0 else np.nan)
    cm = [[tp, fn],
          [fp, tn]]
    pred_results['cm'] = pd.DataFrame(cm, index=['act_pos', 'act_neg'], columns=['pred_pos', 'pred_neg'], dtype=int)
    return pred_results


def export_gene_predictions(pred_results, exp_fitness, pred_fitness, pred_status, uniprot_data, exp_mpmf, fname=None):
    """Export gene predictions with additional information.

    Using the structure returned by `confusion_matrix()` a table is generated, indexed by gene id. The table will
    be written to an Excel file, if `fname` is provided.
    Table contains additional data, extracted from information provided in the parameters.

    For gene essentiality analysis, set parameter `exp_fitness` to {}.

    Example: Perform single gene deletion simulation (using gurobipy interface) and export prediction results.
    `keio_ess` and `keio_red` hold lists of genes that are considered essential/redundant for selected condition.
    `df_mpmf` contains proteomics data for reference. Uniprot data is collected for the organism in question.

    .. code-block:: python

        from f2xba.uniprot.uniprot_data import UniprotData

        uniprot_data = UniprotData(83333, 'data_refs')

        eo = EcmOptimization('iML1515_GECKO.xml')
        eo.medium = {rid: 1000.0 for rid in lb_medium}
        df_sgko = eo.single_gene_deletion()

        act_classification = {gene: False for gene in keio_red}
        act_classification.update({gene: True for gene in keio_ess})
        pred_classification = (df_sgko['fitness'] < 0.05).to_dict()
        pred = confusion_matrix(act_classification, pred_classification)

        pred_fitness = df_sgko['fitness'].to_dict()
        pred_status = df_sgko['status'].to_dict()
        exp_mpmf = df_mpmf['LB'].to_dict()
        fname = 'essentiality_predictions.xlsx'
        df_predictions = export_gene_predictions(pred, {}, pred_fitness, pred_status, uniprot_data, exp_mpmf, fname)

    :param dict pred_results: SGKO prediction results generated by confusion_matrix()
    :param dict(str, float) exp_fitness: fitness data from experiment, if available, otherwise {}
    :param dict(str, float) pred_fitness: fitness data determined from SGKO analysis
    :param dict(str, str) pred_status: optimization status of SGKO predictions
    :param uniprot_data: instance containing UniProt protein data for given model/organism
    :type uniprot_data: :class:`UniprotData`
    :param dict(str, float) exp_mpmf: experimental values of protein mass fractions in mg/g
    :param str fname: (optional) Excel file name of spreadsheet with`.xlsx`
    :return: table with detailed prediction data
    :rdata: pandas.DataFrame
    """
    cols = ['gene_name', 'prediction', 'pred_fitness', 'pred_status', 'exp_fitness', 'mpmf',
            'uid', 'aa_len', 'description', 'go_processes']
    data = {}
    for pred_cat in ['fp', 'fn', 'tp', 'tn']:
        for gene in pred_results[pred_cat]:
            uid = uniprot_data.locus2uid[gene]
            prot = uniprot_data.proteins[uid]
            data[gene] = [prot.gene_name, pred_cat, pred_fitness.get(gene), pred_status.get(gene),
                          exp_fitness.get(gene),
                          exp_mpmf.get(gene), uid, prot.length, prot.protein_name, '; '.join(prot.go_processes)]
    df_predictions = pd.DataFrame(data.values(), index=list(data), columns=cols)
    df_predictions.sort_index(inplace=True)
    df_predictions.index.name = 'gene'

    # write predictions results to file
    if fname:
        with pd.ExcelWriter(fname) as writer:
            df_predictions.to_excel(writer, sheet_name='SGKO predictions')
            print(f'prediction results written to {fname}')
    return df_predictions



"""Implementation of TdMReaction class.

holding thermodynamics data for a reaction
based on pyTFA thermo data

Peter Schubert, HHU Duesseldorf, Octobert 2023
"""
import re
import numpy as np
from collections import defaultdict
import f2xba.prefixes as pf
from .td_constants import CPD_WATER, CPD_PROTON, RT, FARADAY, DEFAULT_DRG_ERROR


class TdReaction:

    def __init__(self, r, cid2nh, cid2charge):
        """Instantiate thermodynamics reaction.

        Note: for strings we convert the type from numpy.str_ to str

        :param r: associated model reaction object
        :type r: :class:`SbmlReaction`
        :param dict cid2nh: net number of H atoms, incl. protons per compartment
        :param dict cid2charge: net charge per compartment
        """
        self.id = r.id
        self.r = r
        self.r_metabolites = {}
        for reactant_side, sign in {'reactants': -1.0, 'products': 1.0}.items():
            for sid, stoic in getattr(r, reactant_side).items():
                if re.match(pf.C_, sid) is None:
                    self.r_metabolites[sid] = sign * stoic

        self.tx_nhs = self._get_tx_items(cid2nh, 'nh')
        self.tx_charges = self._get_tx_items(cid2charge, 'charge')
        self.drg0_tr = 0.0
        self.sum_dfg0_tr = 0.0
        self.sum_nh_adjust = 0.0
        self.sum_el_work = 0.0
        self.drg_error = 0.0
        self.rtlnq = 0.0

    # updated determination of transported h - atoms and charges
    @staticmethod
    def _get_tx_items(cid2item, item):
        """Determine detailed flow of transported H atoms and charges.

        Net transported charges / H-atoms are reported from -> to compartmets
        Check that all charges/H-atoms are accounted for.

        Support transport of charges or H-atoms across more than single membrane.
        E.g. Export compound from periplasm to external medium driven by proton gradient
        across internal membrane (i.e. proton import into cytosol)

        :param str item: 'nh' or 'charge
        :return: number H or number charge transfered from/to compartments
        :rtype: list of dict (keys: from, to, nh or charge)
        """
        from_cids = [cid for cid, item in cid2item.items() if item <= 0.0]
        to_cids = [cid for cid, item in cid2item.items() if item > 0.0]
        tx_items = []
        for from_cid in from_cids:
            from_item = cid2item[from_cid]
            for to_cid in to_cids:
                to_item = cid2item[to_cid]
                if from_item + to_item > 0:
                    to_item = -from_item
                tx_items.append({'from': from_cid, 'to': to_cid, item: to_item})
                from_item += to_item
                cid2item[to_cid] -= to_item
            assert from_item == 0
            cid2item[from_cid] = from_item

        return tx_items

    def get_gibbs_reaction_energy_std_tr(self, td_reactants, td_compartments):
        """Calculate standard transformed Gibbs energy of reaction.

        Supporting non-transport ond transport reactions.

        Using formulation of Jol, 2010, equation 6, aligned with formalism used in equilibrator (based on
        Haraldsdottir et al. 2012):

            ∆gG'˚ = ∑ v_i ∆fG'˚_i
                    - RT ln(10) ∑ NH_od (-pH_s + pH_d)
                    + F ∆pot_sd ∑ charge_sd
            v_i: stoichiometry of reactant
            ∆fG'˚_i: standard transformed Gibbs energy of formation of reactant i
            NH_sd: net H atoms, incl. protons transported from source to destination compartment
            pH_s, pH_d: pH values in source and destination compartments
            ∆pot_sd: difference in electrical potential between source and destination compartments (pot_d - pot_s)
            charge_sd: net charge transported from source to destination compartment

        support transport reactions across more than singel membrane

        :param dict td_reactants: thermodynamic reactants
        :param dict td_compartments: thermodynamic compartments
        :returns: standard transformed Gibbs energy of reaction in kJ/mol
        :rtype: float
        """

        # ∑ v_i ∆fG'˚_i
        self.sum_dfg0_tr = 0.0
        for sid, stoic in self.r_metabolites.items():
            td_rs = td_reactants[sid]
            if td_rs.td_metabolite != CPD_PROTON:
                self.sum_dfg0_tr += stoic * td_rs.dfg0_tr

        # - RT ln(10) ∑ NH_sd (-pH_s + pH_d), for net H atoms, including protons, transported
        self.sum_nh_adjust = 0.0
        for data in self.tx_nhs:
            dph = td_compartments[data['to']].ph - td_compartments[data['from']].ph
            self.sum_nh_adjust -= RT * np.log(10) * data['nh'] * dph

        # + F ∑ ∆pot_sd     charge_sd, net charge transorted from source to destination
        self.sum_el_work = 0.0
        for data in self.tx_charges:
            dpot = td_compartments[data['from']].membrane_pots[data['to']]
            self.sum_el_work += FARADAY * dpot * data['charge']

        self.drg0_tr = self.sum_dfg0_tr + self.sum_nh_adjust + self.sum_el_work

        return self.drg0_tr

    def get_gibbs_reaction_energy(self, td_reactants, td_compartments, concentrations):
        """Calculate Gibbs transformed energy of reaction for given reactant concentrations.

        Can be used in TfaModel to determine Gibbs energy of reactions.

        ∆rG' = ∆rG'˚ + RT ∑ v_i ln(B_i)   (excluding water and protons)

        :param dict td_reactants: thermodynamic reactants
        :param dict td_compartments: thermodynamic compartments
        :param dict concentrations: reactant concentrations in mol/l
        :return: Gibbs transformed energy of reaction for given reactant concentrations
        :rtype: float
        """
        # + RT ∑ v_i ln(B_i) (excluding water and protons)
        self.rtlnq = 0.0
        for sid, stoic in self.r_metabolites.items():
            td_rs = td_reactants[sid]
            if td_rs.td_metabolite not in {CPD_PROTON, CPD_WATER}:
                if sid in concentrations:
                    self.rtlnq += RT * stoic * np.log(concentrations[sid])
                else:
                    print(f'Reactant concentration for {sid} not provided/considered')

        drg0_tr = self.get_gibbs_reaction_energy_std_tr(td_reactants, td_compartments)
        return drg0_tr + self.rtlnq

    def get_gibbs_reaction_error(self, td_reactants, td_cues):
        """Calculate ∆Gr error related to ∆Gf of metabolites using group contribution method.

        Error formulation as per pyTFA: The actual cues being modified by the reaction are
        considered for the error estimation.
        drg_error == sqrt(∑ (cue_est_error * stoic)^2)

        :param dict td_reactants: thermodynamic reactants
        :param dict td_cues: thermodynamic cues data related ot model reactants
        :returns: Error in determining Gibbs energy of reaction in kJ/mol
        :rtype: float
        """
        srefs = {}
        for reactant_side, sign in {'reactants': -1.0, 'products': 1.0}.items():
            for sid, stoic in getattr(self.r, reactant_side).items():
                if re.match(pf.C_, sid) is None:
                    srefs[sid] = stoic * sign

        # for the reaction, we identify the actual cues being converted
        cues_balance = defaultdict(float)
        for sid, stoic in srefs.items():
            td_m = td_reactants[sid].td_m
            for cue_id, species_cue_stoic in td_m.struct_cues.items():
                cues_balance[cue_id] += stoic * species_cue_stoic
        cues_unbalance = {cue_id: balance for cue_id, balance in cues_balance.items() if balance != 0.0}

        total_cues_dg_error = 0.0  # pyTFA calculation sqrt of sum of squared errors
        for cue_id, reaction_cues_stoic in cues_unbalance.items():
            cue_data = td_cues[cue_id]
            total_cues_dg_error += (cue_data.error * reaction_cues_stoic) ** 2
        self.drg_error = np.sqrt(total_cues_dg_error)

        if self.drg_error == 0.0:
            self.drg_error = DEFAULT_DRG_ERROR

        return self.drg_error

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        :param str attribute: attribute name
        :param value: value to be configured
        :type value: str, float, bool
        """
        if hasattr(self, attribute):
            setattr(self, attribute, value)
        else:
            print(f'unknown TD metabolite attribute {attribute}')



"""Implementation of TfaModel class.

Extend the XbaModel to become a thermodynamics constraint model.
The implementation of TFA is based on pyTFA (Salvy et al., 2019).
Thermodynamics data is converted to kJ/mol.

Peter Schubert, HHU Duesseldorf, October 2023
"""

import os
import re

import numpy as np
import pandas as pd
from collections import defaultdict
import pickle
import zlib

from .td_constants import CPD_PROTON, CPD_WATER, RT, MAX_DRG, KJ_PER_KCAL
from .td_compartment import TdCompartment
from .td_metabolite import TdMetabolite
from .td_cue import TdCue
from .td_reaction import TdReaction
from .td_reactant import TdReactant
from ..utils.calc_mw import extract_atoms
from ..utils.mapping_utils import load_parameter_file, construct_metabolic_reactions_string
from ..xba_model.fbc_objective import FbcObjective
import f2xba.prefixes as pf


class TfaModel:
    """Extend a XbaModel instance by adding thermodynamic constraints.

    The extended XbaModel can be used to create a TFA model, thermodynamics enabled
    enzyme constraint (e.g. TGECKO) and resource balance constraint (TRBA) models.

    Optimization of models might produce infeasible results, when adding thermodynamic
    constraints. In such cases, parameter relaxation of the TD constrain model might be required.
    Relaxation would adjust bounds of ∆rG'˚ (standard transformed Gibbs energy of reaction)
    optimization variables. The adjusted bounds can be included in the TFA configuration file.
    See tutorial.

    The implementation is based on Python package pyTFA (Salvy et al., 2019).

    Ref: Salvy, P., Fengos, G., Ataman, M., Pathier, T., Soh, K. C., & Hatzimanikatis, V. (2019).
    pyTFA and matTFA: a Python package and a Matlab toolbox for Thermodynamics-based Flux Analysis.
    Bioinformatics, 35(1), 167-169. https://doi.org/10.1093/bioinformatics/bty499

    Example 1: Create a TFA model by extending an existing genome scale metabolic model.
    The spreadsheet file `tfa_parameters.xlsx` contains configuration data. See tutorials.

    .. code-block:: python

        xba_model = XbaModel('iML1515.xml')
        xba_model.configure()

        tfa_model = TfaModel(xba_model)
        tfa_model.configure('tfa_parameters.xlsx')
        if tfa_model.validate():
            tfa_model.export('iML1515_TFA.xml')

    Example 2: Create a TGECKO model. The spreadsheet files `xba_parameters.xlsx`, `tfa_parameters.xlsx`
    and `ecm_parameters.xlsx` contain configuration data.

    .. code-block:: python

        xba_model = XbaModel('iML1515.xml')
        xba_model.configure('xba_parameters.xlsx')

        tfa_model = TfaModel(xba_model)
        tfa_model.configure('tfa_parameters.xlsx')

        ec_model = EcModel(xba_model)
        ec_model.configure('ecm_parameters.xlsx')
        ec_model.export('iML1515_TGECKO.xml')

    """

    def __init__(self, xba_model):
        """Instantiate the TfaModel instance.

        :param xba_model: a reference to the XbaModel instance
        :type xba_model: :class:`XbaModel`
        """
        self.model = xba_model
        """Reference to the XbaModel instance."""

        self.td_params = {}
        """General parameters extracted from TFA configuration file, sheet `general`."""

        self.td_compartments = {}
        """TD compartments as per TFA configuration file."""

        self.td_metabolites = {}
        """TD metabolites extracted from the TD database."""

        self.td_cues = {}
        """TD data of TD cues (components of TD species), extracted from the TD database."""

        self.td_reactants = {}
        """TD data related to model species."""

        self.td_reactions = {}
        """TD data related to model reactions."""

    def configure(self, tfa_config_data):
        """Configure the TfaModel instance with information provided in the TFA configuration data (file or dict).

        f2xba can determine standard transformed Gibbs reaction energies using a formulation aligned with pyTFA.
        It is however possbile to supply standard transformed Gibbs reaction energies and related errors, which
        have been determined by different formulations, e.g. using equilibrator.

        User provided standard transformed Gibbs reaction energies can be provided in an Excel spreadsheet
        with file name provided in the optional parameter 'drg0trs_fname' in the 'general' table of the
        TFA configuration data. In this case a thermo database needs not to be provided.

        The table 'td_compartments' is mandatory and tables 'mid2tdmid', 'modify_thermo_data',
        'modify_drg0_bounds' are optional

        .. code-block:: python

                xba_model = XbaModel('iML1515.xml')
                xba_model.configure()

                tfa_model = TfaModel(xba_model)
                tfa_model.configure('tfa_parameters.xlsx')

        :param str or dict tfa_config_data: filename of TFA configuration file (.xlsx) or dict with config data.
        :return: success status of operation
        :rtype: bool
        """

        if type(tfa_config_data) is str:
            sheet_names = ['general', 'td_compartments', 'mid2tdmid',
                           'modify_thermo_data', 'modify_drg0_bounds']
            tfa_params = load_parameter_file(tfa_config_data, sheet_names)
            if 'general' not in tfa_params.keys() or 'td_compartments' not in tfa_params.keys():
                print(f'mandatory tables "general" and "td_compartments" must be provided')
                raise ValueError
        elif type(tfa_config_data) is dict:
            tfa_params = tfa_config_data
        else:
            print('parmater "tfa_config_data" is neither a valid file name nor a dict with config data')
            raise ValueError
        self.td_params = tfa_params['general']['value'].to_dict()

        self.td_compartments = {cid: TdCompartment(str(cid), row)
                                for cid, row in tfa_params['td_compartments'].iterrows()}

        if 'drg0trs_fname' not in self.td_params:
            all_td_metabs, all_td_cues = self._load_thermo_db(self.td_params['thermo_data_fname'], tfa_params)
            self._create_td_metabolites(all_td_metabs, all_td_cues, tfa_params)
            self._create_td_cues(all_td_cues)
            self._create_td_reactants()
            self._create_td_reactions()
        else:
            self._create_td_reactions_from_file(self.td_params['drg0trs_fname'])

        # adding TFA related constraints and variables
        self._add_tfa_constraints()
        self._add_tfa_variables()

        # in case relaxed parameters are already available, implement them
        if 'modify_drg0_bounds' in tfa_params:
            dgr0_vids = []
            for var_id in tfa_params['modify_drg0_bounds'].index:
                dgr0_vids.append(var_id)
            self._modify_drg0_bounds(tfa_params['modify_drg0_bounds'].loc[dgr0_vids], remove_slack=False)

        # modify some model attributs and create L3V2 SBML model
        self.model.model_attrs['id'] = self.model.model_attrs.get('id', 'model') + '_TFA'
        if 'name' in self.model.model_attrs:
            self.model.model_attrs['name'] = f'TFA model of ' + self.model.model_attrs['name']
        self.model.sbml_container['level'] = 3
        self.model.sbml_container['version'] = 2

        self.print_td_stats()
        self.model.print_size()
        return True

    def print_td_stats(self):
        """Print TD statistics on species/reactions.

        :meta private:
        """
        var_ids = set(self.model.reactions)
        sids = []
        td_metab = []
        mids = set()
        td_mids = set()
        for constr_id in self.model.species:
            if re.match(pf.M_, constr_id):
                sidx = re.sub(f'^{pf.M_}', '', constr_id)
                midx = sidx.rsplit('_', 1)[0]
                sids.append(sidx)
                mids.add(midx)
                if pf.V_LC_ + sidx in var_ids:
                    td_metab.append(sidx)
                    td_mids.add(midx)

        orig_rids = set()
        td_rids = set()
        for var_id, r in self.model.reactions.items():
            if re.match(pf.R_, var_id):
                if r.kind in ['transporter', 'metabolic']:
                    ridx = re.sub(f'^{pf.R_}', '', r.orig_rid)
                    orig_rids.add(ridx)
                    if pf.V_DRG_ + ridx in var_ids:
                        td_rids.add(ridx)
        print(f'{len(td_metab)} species ({len(td_mids)} metabolites) with TD data from total {len(sids)} ({len(mids)})')
        print(f'{len(td_rids)} metabolic/transporter reactions with TD data from total {len(orig_rids)}')

    def validate(self):
        """Validate compliance with SBML standards, including units configuration.

        Validation is an optional task taking time. Validation could be skipped once model
        configuration is stable.

        Information on non-compliance is printed. Details are written to `tmp/tmp.txt`.
        In case of an unsuccessful validation, it is recommended to review `tmp/tmp.txt` and
        improve on the model configuration.

        Example: Ensure compliance with SBML standards for a TfaModel instance prior to its export to file.

        .. code-block:: python

            if tfa_model.validate():
                tfa_model.export('iML1515_TFA.xml')

        :return: success status
        :rtype: bool
        """
        # remove model components that are not used in the model and update SBML groups
        self.model.clean()

        return self.model.validate()

    def export(self, fname):
        """Export TfaModel to SBML encoded file (.xml) or spreadsheet (.xlsx).

        The spreadsheet (.xlsx) is helpful to inspect model configuration.

        Example: Export TfaModel to SBML encoded file and to spreadsheet format.

        .. code-block:: python

            if tfa_model.validate():
                tfa_model.export('iML1515_TFA.xml')
                tfa_model.export('iML1515_TFA.xlsx')

        :param str fname: filename with extension '.xml' or '.xlsx'
        :return: success status
        :rtype: bool
        """
        # remove model components that are not used in the model and update SBML groups
        self.model.clean()
        self.model.print_size()

        return self.model.export(fname)

    def export_drg0_tr(self, fname):
        """Export standard transformed Gibbs reaction energies and errors to file (Excel spreadsheet).

        Export Gibbs reaction data from a TFA model instance, after values have been determined
        by f2xba formulation (based on pyTFA)

        The file can be used as a template to supply standard transformed Gibbs reaction
        energies and errors calculated by different methods.

        .. code-block:: python

            xba_model = XbaModel('iML1515.xml')
            xba_model.configure('xba_parameters.xlsx')
            tfa_model = TfaModel(xba_model)
            tfa_model.configure('tfa_parameters.xlsx')
            tfa_model.export_drg0_tr('drgotrs.xlsx')

        :param str fname: filename with extension '.xlsx'
        """

        cols = ['rid', 'drg0tr_kJ_per_mol', 'error_kJ_per_mol', 'info_reaction_string']
        drg0tr_data = []
        for rid, data in self.td_reactions.items():
            drg0tr_data.append([rid, data.drg0_tr, data.drg_error, construct_metabolic_reactions_string(data.r)])
        df = pd.DataFrame(drg0tr_data, columns=cols).set_index('rid').sort_index()

        with pd.ExcelWriter(fname) as writer:
            df.to_excel(writer, sheet_name='drg0tr')
            print(f'{len(df)} standard transformed Gibbs reaction energies exported to', fname)

    @staticmethod
    def _load_thermo_db(thermo_db_fname, tfa_params):
        """Load TD database and configure TD data based on TFA configuration file.

        Energy units, if provided in kcal/mol are converted to kJ/mol

        TD database must be formatted as per TD database used by pyTFA (Salvy et al., 2019), see
        https://github.com/EPFL-LCSB/pytfa/raw/refs/heads/master/data/thermo_data.thermodb

        - Load TD database from file (format must similar to pyTFA thermo_data.thermodb).
        - update TD database record attributes based on TFA configuration table 'modify_thermo_data'.
        - convert energy value from kcal/mol to kJ/mol, if units the TD database are in kcal/mol.

        :param str thermo_db_fname: file name of thermodynamics database
        :param dict(str, pandas.DataFrame) tfa_params: configuration tables from TFA configuration file
        :return all_thermo_metabolites and all_thermo_cues
        :rtype dict, dict
        """
        # load thermo dynamic data from file and correct some errors
        if not os.path.exists(thermo_db_fname):
            print(f'{thermo_db_fname} does not exist')
            raise FileNotFoundError
        with open(thermo_db_fname, 'rb') as file:
            all_thermo_data = pickle.loads(zlib.decompress(file.read()))

        # update TD data records based on TFA configuration table 'modify_thermo_data'
        if 'modify_thermo_data' in tfa_params:
            count = 0
            for td_metab, row in tfa_params['modify_thermo_data'].iterrows():
                if td_metab in all_thermo_data['metabolites']:
                    if row['attribute'] in all_thermo_data['metabolites'][td_metab]:
                        value = row['value']
                        # handle lists, e.g. pKa values
                        match = re.match(r'\[(.*)]$', str(value))
                        if match:
                            value = [float(val) for val in match.group(1).split(',') if len(val.strip()) > 0]
                        all_thermo_data['metabolites'][td_metab][row['attribute']] = value
                        count += 1
                    else:
                        print(f'{td_metab} has no attribute {row["attribute"]} in TD database')
                else:
                    print(f'{td_metab} id not found in TD database (check table "modify_thermo_data")')
            print(f'{count:4d} thermo data attributes updated')

        # flag invalid energy values and convert to energy values to units of kJ/mol
        conv_factor = KJ_PER_KCAL if all_thermo_data['units'] == 'kcal/mol' else 1.0
        for td_mid in all_thermo_data['metabolites'].keys():
            data = all_thermo_data['metabolites'][td_mid]
            data['deltaGf_std'] = data['deltaGf_std'] * conv_factor if abs(data['deltaGf_std']) < 1e6 else None
            data['deltaGf_err'] = data['deltaGf_err'] * conv_factor if abs(data['deltaGf_err']) < 1e6 else None
        for td_cue in all_thermo_data['cues'].keys():
            data = all_thermo_data['cues'][td_cue]
            data['energy'] = data['energy'] * conv_factor if abs(data['energy']) < 1e3 else None
            data['error'] = data['error'] * conv_factor if abs(data['error']) < 1e3 else None

        return all_thermo_data['metabolites'], all_thermo_data['cues']

    def _check_compatible_thermo_data(self, sid, metabolite_td_data):
        """Check compatibility between the model species corresponding thermo data record.

        For the check to pass, the TD record must have a valid ∆Gf° and be error free.
        Further, the formula and chemical charge, if configured on the model species, must
        be compatible with the formula and chemical charge configured in the TD record. The number
        of protons and corresponding charges are allowed to differ.

        :meta private:
        :param str sid: species id in the model
        :param dict metabolite_td_data: metabolite record extracted from TD data
        :return: success/failure of test
        :rtype: bool
        """
        valid = False
        # only accept corresponding TD record with valid Gibbs free energy of formation
        if metabolite_td_data['deltaGf_std'] < 1e6 and metabolite_td_data['deltaGf_err'] < 1e6:
            s = self.model.species[sid]

            # check difference of chemical formula of model species vs. selected TD metabolite
            td_atom_delta = {}
            if hasattr(s, 'formula') and type(s.formula) is str:
                s_atoms = {atom: count for atom, count in extract_atoms(s.formula)}
                # Note: strings loaded via pickle from thermo db have type numpy.str_
                td_atoms = {atom: count for atom, count in extract_atoms(str(metabolite_td_data['formula']))}
                atoms = set(s_atoms.keys()).union(set(td_atoms.keys()))
                for atom in atoms:
                    delta = td_atoms.get(atom, 0) - s_atoms.get(atom, 0)
                    if delta != 0:
                        td_atom_delta[atom] = delta

            td_charge_delta = 999
            if hasattr(s, 'charge') and (isinstance(s.charge, float) or isinstance(s.charge, int)):
                td_charge_delta = metabolite_td_data['charge_std'] - s.charge

            # OK if chemical formula and charges match between model species and TD metabolite
            if len(td_atom_delta) == 0 and td_charge_delta == 0:
                valid = True
            # also OK if charge difference is based on different protonation state
            elif (len(td_atom_delta) == 1) and ('H' in td_atom_delta) and (td_atom_delta['H'] == td_charge_delta):
                valid = True
        return valid

    def _create_td_metabolites(self, all_td_metabs, all_td_cues, tfa_params):
        """Create TD metabolites related to species defined in the model.

        We make used of SEED refs in species model annotation and species name.
        We configure 'td_metab' attribute on model species to corresponding TD species,
          alternatively, it is set to None
        only select TD species, that contain a valid formula

        Assignment of a td_metab is done in steps:
            1. we can set a specific td_metab in mid2tdmid
                - for this a metabolite id is extracted from the model species id using 'mid_regex_pattern'
            2. next, we try finding a matching metabolite name in thermo data
                - name matching is in lower case
            3. finaly we use one seed id from model annotation that exists in thermo data
                - first seed id in a sorted list of model seed references
                - Note: thermo data ids are based on seed ids, however they are not fully compliant to Model.Seed

            Finaly, we check that selected TD record has a valid ∆Gf° and is error free. In addition, the
            chemical formula and electrical charge, if configured on the model species, must be compatible
            with the TD data record. Only exception is a difference in protonation state and corresponding
            charge.

        If no valid TD record can be found for a model species, reactions using the species cannot
        be configured with TD constraints.

        :param dict(str, dict) all_td_metabs: reference to all TD metabolite records in the TD database
        :param dict(str, dict) all_td_cues: reference to all TD cues records in the TD database
        :param dict(str, pandas.DataFrame) tfa_params: configuration tables from TFA configuration file
        """
        # get configuration data from TFA configuration file
        mid2tdmid = tfa_params['mid2tdmid']['tdmid'].to_dict() if 'mid2tdmid' in tfa_params else {}
        mid_regex_pattern = tfa_params['general'].get('mid_regex_pattern', r'^M_(\w+)_\w+$')

        # for metabolite name mapping, we create a dict of names in thermo data to respective seed id
        lcname2seed_id = {}
        valid_td_metabolites = set()
        for seed_id, data in all_td_metabs.items():
            if data['formula'] not in ['NA', 'NULL'] and data['deltaGf_std'] is not None:
                valid_td_metabolites.add(seed_id)
                lcname2seed_id[data['name'].lower()] = seed_id
                for name in data['other_names']:
                    if len(name) > 0:
                        lcname2seed_id[name.lower()] = seed_id

        supported_species = 0
        td_metabs = set()
        for sid, s in self.model.species.items():
            selected_td_metabolite = None

            # check manual mapping based on configuration file
            m = re.match(mid_regex_pattern, sid)
            if m:
                mid = m.group(1)
                if (mid in mid2tdmid) and (mid2tdmid[mid] in valid_td_metabolites):
                    selected_td_metabolite = mid2tdmid[mid]

            # automated mapping
            if selected_td_metabolite is None:
                lcname = s.name.lower()
                if lcname in lcname2seed_id:
                    # mapping based on model species name
                    selected_td_metabolite = lcname2seed_id[lcname]
                else:
                    # mapping based on model species Miriam Annotation - seed.compound reference
                    eligible_td_metabolites = set(s.seed_refs).intersection(valid_td_metabolites)
                    if len(eligible_td_metabolites) > 0:
                        selected_td_metabolite = sorted(eligible_td_metabolites)[0]

            # ensure that chemical formula and electrical charges are compatible
            if selected_td_metabolite is not None:
                if self._check_compatible_thermo_data(sid, all_td_metabs[selected_td_metabolite]):
                    td_metabs.add(selected_td_metabolite)
                    s.modify_attribute('td_metabolite', str(selected_td_metabolite))
                    supported_species += 1

        self.td_metabolites = {str(td_metab): TdMetabolite(td_metab, all_td_metabs[td_metab], all_td_cues)
                               for td_metab in td_metabs}

        print(f'{len(self.td_metabolites):4d} metabolites with TD data covering {supported_species} model species; '
              f'({ len(self.model.species) - supported_species} not supported)')

    def _create_td_cues(self, all_td_cues):
        """Create the TD cues data for cues used requried by species defined in the model.

        td_cues are used to determine protonation state of species and to calculate erros.
        Note: some cues defined for metabolites in the TD database may reference undefined cues

        :param dict all_td_cues: cues data from thermo database
        """
        # create td_cues - selecting required cues only
        for td_sdata in self.td_metabolites.values():
            to_del = set()
            for cue_id in td_sdata.struct_cues:
                if cue_id not in self.td_cues:
                    if cue_id in all_td_cues:
                        self.td_cues[cue_id] = TdCue(cue_id, all_td_cues[cue_id])
                    else:
                        to_del.add(cue_id)
            # remove unsupported cues in td_species, e.g. 'NO'
            for cue_id in to_del:
                del td_sdata.struct_cues[cue_id]

    def _create_td_reactants(self):
        """Create TD reactants directly related to individual model species.

        Calulate transformed Gibbs energy of formation for TD reactant based on isomer group thermodynamics.

        For each species linked to a TD metabolite, create a TD reactant with compartment
        specific parameters. Calculate ∆fG'˚ subject to compartment pH and ionic strength
        and related values.
        """
        for sid, s in self.model.species.items():
            if getattr(s, 'td_metabolite', None) is not None:
                td_metabolite = self.td_metabolites[s.td_metabolite]
                td_rs = TdReactant(s, td_metabolite, self.td_compartments[s.compartment])
                td_rs.get_gibbs_formation_energy_std_tr()
                self.td_reactants[sid] = td_rs

    def _create_td_reactions(self):
        """Create TD reactions for model reactions where TD data is fully available.

        Calculate transformed Gibbs energy of reaction from based ∆fG'˚ of reactants,
        H atom adjustement for transported metabolites and electrical energy term for
        transported charges.

        TD reactions are created when all reactants/products can be linked to a valid TD species.
        Linking is via td_metabolite attribute configured in model species

        We remove transport reactions for water. Water can have different ∆fG'˚
        in different compartments and there would be no way to reverse reaction
        directionality, as water is not included in reaction quotient.
        """
        total_reactions = 0
        for rid, r in self.model.reactions.items():
            if re.match(pf.V_, rid) is None and r.kind in ['metabolic', 'transporter']:
                total_reactions += 1
                if self.model.is_charge_balanced(r) and self.model.is_atom_balanced(r):
                    # td_metabs is a dict with td_metabolties (identified by seed_ids) and
                    #  corresponding species ids in reactants and products side of the reaction
                    td_metabs = defaultdict(dict)
                    cid2nh = defaultdict(float)
                    cid2charge = defaultdict(float)
                    valid_td_data = True
                    for reactant_side, sign in {'reactants': -1.0, 'products': 1.0}.items():
                        for sid, stoic in getattr(r, reactant_side).items():
                            if re.match(pf.C_, sid) is None:
                                if sid in self.td_reactants:
                                    td_rs = self.td_reactants[sid]
                                    s = self.model.species[sid]
                                    td_metabs[td_rs.td_metabolite][reactant_side] = sid
                                    cid2nh[s.compartment] += sign * stoic * td_rs.nh_species
                                    cid2charge[s.compartment] += sign * stoic * s.charge
                                else:
                                    valid_td_data = False
                    if valid_td_data:
                        valid_td_data = (sum(cid2nh.values()) == sum(cid2charge.values()))

                    # create no td_reaction for pure water transport across membranes
                    water_transport = len(td_metabs) == 1 and CPD_WATER in td_metabs

                    if valid_td_data and not water_transport:
                        td_r = TdReaction(r, dict(cid2nh), dict(cid2charge))
                        td_r.get_gibbs_reaction_energy_std_tr(self.td_reactants, self.td_compartments)
                        td_r.get_gibbs_reaction_error(self.td_reactants, self.td_cues)
                        self.td_reactions[rid] = td_r

        not_supported = total_reactions - len(self.td_reactions)
        print(f'{len(self.td_reactions):4d} reactions supported by TD data; ({not_supported} not supported)')

    def _create_td_reactions_from_file(self, drg0trs_fname):
        """Create TD reactions for user provided Gibbs reaction data.

        User can provide standard transformed Gibbs reaction energies and errors, e.g. determined
        by equilibrator application.

        The file contains a single table with colums 'rid', '

        :param str drg0trs_fname: file name of user supplied Gibbs reaction energies (.xlsx)
        """
       # load thermo dynamic data from file and correct some errors
        if not os.path.exists(drg0trs_fname):
            print(f'{drg0trs_fname} does not exist')
            raise FileNotFoundError

        drg0trs = load_parameter_file(drg0trs_fname)
        df_drg0trs = drg0trs[list(drg0trs.keys())[0]]

        if 'drg0tr_kJ_per_mol' not in df_drg0trs.columns or 'error_kJ_per_mol' not in df_drg0trs.columns:
            print(f'table in {drg0trs_fname} requires mandatory columns "drg0tr_kJ_per_mol" and "error_kJ_per_mol".')
            raise AttributeError

        rids_not_found = []
        for rid, data in df_drg0trs.iterrows():
            if rid in self.model.reactions:
                r = self.model.reactions[rid]
                cid2nh = defaultdict(float)
                cid2charge = defaultdict(float)
                for reactant_side, sign in {'reactants': -1.0, 'products': 1.0}.items():
                    for sid, stoic in getattr(r, reactant_side).items():
                        if re.match(pf.C_, sid) is None:
                            if sid in self.model.species:
                                s = self.model.species[sid]
                                nh_species = {atom: count for atom, count in extract_atoms(s.formula)}.get('H', 0)
                                cid2nh[s.compartment] += sign * stoic * nh_species
                                cid2charge[s.compartment] += sign * stoic * s.charge
                td_r = TdReaction(r, dict(cid2nh), dict(cid2charge))
                td_r.drg0_tr = data['drg0tr_kJ_per_mol']
                td_r.drg_error = data['error_kJ_per_mol']
                self.td_reactions[rid] = td_r
            else:
                rids_not_found.append(rid)
        print(
            f'{len(self.td_reactions):4d} td_reactions created from file; ({len(rids_not_found)} reactions not found)')

    def _add_tfa_constraints(self):
        """Add TFA constraints to model as pseudo species.

        Note: irreversible reactions with TD data get not split
        and some constraints are not required
        """
        constr2name_rev = {pf.C_DRG_: "∆rG'",
                           pf.C_SU_: 'simultaneous use',
                           pf.C_GFC_: "∆rG' forward coupling",
                           pf.C_GRC_: "∆rG' reverse coupling",
                           pf.C_FFC_: 'flux forward coupling',
                           pf.C_FRC_: 'flux reverse coupling'}
        constr2name_irr = {pf.C_DRG_: "∆rG'",
                           pf.C_GFC_: "∆rG' forward coupling",
                           pf.C_FFC_: 'flux forward coupling'}

        pseudo_sids = {}
        for rid, td_r in self.td_reactions.items():
            ridx = re.sub(f'^{pf.R_}', '', rid)
            constr2name = constr2name_rev if td_r.r.reversible else constr2name_irr
            for prefix, name in constr2name.items():
                pseudo_sids[f'{prefix}{ridx}'] = [f'{name} for {rid}', 'c', False, False, False]
        cols = ['name', 'compartment', 'hasOnlySubstanceUnits', 'boundaryCondition', 'constant']
        df_add_species = pd.DataFrame(pseudo_sids.values(), index=list(pseudo_sids), columns=cols)
        print(f'{len(df_add_species):4d} constraints to add')

        self.model.add_species(df_add_species)

    def _add_gibbs_reaction_variables(self):
        """Add ∆rG' and ∆rG'˚ variables to the model as pseudo reactions.

        based on pyTFA

        Variable V_DRG0_<rid> holds the value for the standard transformed Gibbs
        energy of reaction ∆rG'˚, which is constructed from standard transformed Gibbs
        energies of formation of participating reactants plus electrical work terms in case
        of charge transport across membranes.
        ∆rG'˚ is implemented as a variable that is allowed to vary between
        the computed value and the estimation error.

        Variable V_DRG_<rid> holds the value for the transformed Gibbs energy of reaction ∆rG',
        which is based on ∆rG'˚ and the reaction quotient, a summation over the
        ln metabolite concentrations.
        ∆rG' has 'unlimited' bounds.
        Constraint C_DRG_<rid> implements the calculation for ∆rG'
        - C_DRG_<rid>: V_DRG_<rid>  = V_DRG0_<rid> + RT ∑ v_i * V_LC_<sid>_i
        V_DRG_<rid> is  connected via coupling constraints C_GFC_<rid> and C_GRC_<rid>
        to forward V_FU_<rid> and reverse V_RU_<rid> use variables.
        """
        var2name = {'DRG': 'transformed Gibbs energy of reaction',
                    'DRG0': 'standard transformed Gibbs energy of reaction'}

        drg_lb_pid = self.model.get_fbc_bnd_pid(-MAX_DRG, 'kJ_per_mol', 'drg_lb', reuse=False)
        drg_ub_pid = self.model.get_fbc_bnd_pid(MAX_DRG, 'kJ_per_mol', 'drg_ub', reuse=False)

        pseudo_rids = {}
        for rid, td_r in self.td_reactions.items():
            ridx = re.sub(f'^{pf.R_}', '', rid)
            if td_r.r.reversible:
                reactions_str = f'{pf.C_DRG_}{ridx} + {pf.C_GRC_}{ridx} -> {pf.C_GFC_}{ridx}'
            else:
                reactions_str = f'{pf.C_DRG_}{ridx} -> {pf.C_GFC_}{ridx}'
            pseudo_rids[f'{pf.V_DRG_}{ridx}'] = [f'{var2name["DRG"]} for {ridx}', reactions_str,
                                                 drg_lb_pid, drg_ub_pid, 'td_variable', 'continuous']

            drg0_lb = td_r.drg0_tr - td_r.drg_error
            drg0_ub = td_r.drg0_tr + td_r.drg_error
            drg0_lb_pid = self.model.get_fbc_bnd_pid(drg0_lb, 'kJ_per_mol', f'{pf.V_DRG0_}{ridx}_lb', reuse=False)
            drg0_ub_pid = self.model.get_fbc_bnd_pid(drg0_ub, 'kJ_per_mol', f'{pf.V_DRG0_}{ridx}_ub', reuse=False)
            pseudo_rids[f'{pf.V_DRG0_}{ridx}'] = [f'{var2name["DRG0"]} for {ridx}', f'-> {pf.C_DRG_}{ridx}',
                                                  drg0_lb_pid, drg0_ub_pid, 'td_variable', 'continuous']

        cols = ['name', 'reactionString', 'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pseudo_rids.values(), index=list(pseudo_rids), columns=cols)
        print(f"{len(df_add_rids):4d} ∆rG'/∆rG'˚ variables to add")
        self.model.add_reactions(df_add_rids)

    def _add_use_variables(self):
        """Add forward and reverse use variables to the model as pseudo reactions.

        based on pyTFA

        Forward 'V_FU_<rid>' and Reverse 'V_RU_<rid>' use variables couple reaction
        directionality to ∆rG' via several constraints:
            - C_FFC_<rid>: connects forward reaction flux to V_FU_<rid>
                C_FFC_<rid>: R_<rid>_fwd - MAX_FLUX V_FU_<rid> <= 0
                i.e. if forward reaction flux > 0 -> V_FU_<rid> = 1
            - C_SU_<rid>: restricts simultaneous use of 'V_FU_<rid>' and 'V_BU_<rid>'
                C_SU_<rid>: V_FU_<rid> + V_RU_<rid> <= 1
                i.e. if V_FU_<rid> = 1 -> V_RU_<rid> = 0
            - C_FRC_<rid>: connects reverse reaction flux to V_RU_<rid>
                C_FRC_<rid>: R_<rid>_rev - MAX_FLUX V_RU_<rid> <= 0
                i.e. if V_RU_<rid> = 0 -> reverse reaction flux = 0
            - C_GFC_<rid>: couples negative ∆rG' (V_DRG_<rid>) to V_FU_<rid>
                C_GFC_<rid>: V_DRG_<rid> + MAX_DRG V_FU_<rid> <= MAX_DRG - eps
                eps = 1e-8
                i.e. if V_FU_<rid> = 1 -> V_DRG_<rid> < - eps
            - C_GRC_<rid>: couples positive ∆rG' (V_DRG_<rid>) to V_RU_<rid>
                C_GRC_<rid>: - V_DRG_<rid> + MAX_DRG V_RU_<rid> <= MAX_DRG - eps
                i.e. if V_RU_<rid> = 0 -> V_DRG_<rid> unconstraint
                i.e. if V_RU_<rid> = 1 -> V_DRG_<rid> >= eps

        Forward and backward use variables are binary variables (values 0 and 1) and must be
        configured as such in the linear problem before optimization,
        - e.g. in CobraPy: tfa_model.variables['V_FU_<rid>'].type = 'binary'
        Upper constraint bounds must be also configure in the linear problem.
        - e.g. tfa_slack_model.constraints['C_SU_<rid>'].lb = None
        - e.g. tfa_slack_model.constraints['C_SU_<rid>'].ub = 1.0
        """
        var2name = {'FU': 'forward use variable',
                    'RU': 'reverse use variable'}
        fbc_max_abs_flux_val = max(abs(self.model.fbc_flux_range[0]), abs(self.model.fbc_flux_range[1]))

        lb_pid = self.model.get_fbc_bnd_pid(0.0, 'fbc_dimensionless', 'binary_use_vars_lb', reuse=False)
        ub_pid = self.model.get_fbc_bnd_pid(1.0, 'fbc_dimensionless', 'binary_use_vars_ub', reuse=False)

        pseudo_rids = {}
        for rid, td_r in self.td_reactions.items():
            ridx = re.sub(f'^{pf.R_}', '', rid)
            if td_r.r.reversible:
                fu_reaction = (f'{fbc_max_abs_flux_val} {pf.C_FFC_}{ridx} => '
                               f'{MAX_DRG} {pf.C_GFC_}{ridx} + {pf.C_SU_}{ridx}')
                bu_reaction = (f'{fbc_max_abs_flux_val} {pf.C_FRC_}{ridx} => '
                               f'{MAX_DRG} {pf.C_GRC_}{ridx} + {pf.C_SU_}{ridx}')
                pseudo_rids[f'{pf.V_RU_}{ridx}'] = [f'{var2name["RU"]} for {ridx}', bu_reaction,
                                                    lb_pid, ub_pid, 'td_variable', 'binary']
            else:
                fu_reaction = f'{fbc_max_abs_flux_val} {pf.C_FFC_}{ridx} => {MAX_DRG} {pf.C_GFC_}{ridx}'
            pseudo_rids[f'{pf.V_FU_}{ridx}'] = [f'{var2name["FU"]} for {ridx}', fu_reaction,
                                                lb_pid, ub_pid, 'td_variable', 'binary']

        cols = ['name', 'reactionString', 'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pseudo_rids.values(), index=list(pseudo_rids), columns=cols)
        print(f'{len(df_add_rids):4d} forward/reverse use variables to add')
        self.model.add_reactions(df_add_rids)

    def _add_log_concentration_variables(self):
        """Add log10 concentration variables for reactants.

        log10 concentrations  variables V_LC_<sid> hold the concentrations
        of reactants participating in reactions having complete TD data.

        log concentrations are required for calculation of Gibbs energy
        of reactions ∆rG' and are part of the reaction quotient
        The bound of the log concentrations are determined by min/max
        concentration values configured on species.

        Constraint C_DRG_<rid> contains the calculation for ∆rG'
        - C_DRG_<rid>: - V_DRG_<rid> + V_DRG0_<rid> + RT ln(10) ∑ v_i * V_LC_<sid>_i = 0 (kJ/mol)

        Protons (H) are not included in transformed Gibbs energy calculations, as
        proton concentration is held constant by the compartment pH.
        Water (H2O) is generally not included in reaction quotient for reaction in diluted
        aqueous solutions.
        """
        var2name = {'LC': 'log reactant concentration'}

        # first we identify for each model species the reactions (including stoic) it participates in
        lc_variables = defaultdict(dict)
        for rid, td_r in self.td_reactions.items():
            ridx = re.sub(f'^{pf.R_}', '', rid)

            # RT ln(10) ∑ S_i log(C_i) for reactants, excluding water and protons
            for sid, stoic in td_r.r_metabolites.items():
                if sid in self.td_reactants:
                    if self.td_reactants[sid].td_metabolite not in {CPD_PROTON, CPD_WATER}:
                        lc_variables[sid][ridx] = RT * np.log(10) * stoic
                else:
                    # case for user provided Gibbs reaction energies
                    s = self.model.species[sid]
                    if CPD_PROTON not in s.seed_refs and CPD_WATER not in s.seed_refs:
                        lc_variables[sid][ridx] = RT * np.log(10) * stoic

        # generate compartment specific fbc_flux_bound parameters for c_min and c_max
        lc_lb_pids = {}
        lc_ub_pids = {}
        for cid, td_c in self.td_compartments.items():
            lc_lb_pids[cid] = self.model.get_fbc_bnd_pid(np.log10(td_c.c_min), 'fbc_dimensionless',
                                                              f'{pf.V_LC_}{cid}_c_min', reuse=False)
            lc_ub_pids[cid] = self.model.get_fbc_bnd_pid(np.log10(td_c.c_max), 'fbc_dimensionless',
                                                              f'{pf.V_LC_}{cid}_c_max', reuse=False)

        # second we implement variables
        pseudo_rids = {}
        for sid, data in lc_variables.items():
            s = self.model.species[sid]
            sidx = re.sub(f'^{pf.M_}', '', sid)
            reac_str = ' + '.join([f'{-rt_stoic} C_DRG_{ridx}' for ridx, rt_stoic in data.items() if rt_stoic < 0.0])
            prod_str = ' + '.join([f'{rt_stoic} C_DRG_{ridx}' for ridx, rt_stoic in data.items() if rt_stoic > 0.0])
            pseudo_rids[f'{pf.V_LC_}{sidx}'] = [f'{var2name["LC"]} of {sid}', f'{reac_str} -> {prod_str}',
                                                lc_lb_pids[s.compartment], lc_ub_pids[s.compartment],
                                                'td_variable', 'continuous']

        cols = ['name', 'reactionString', 'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pseudo_rids.values(), index=list(pseudo_rids), columns=cols)
        print(f'{len(df_add_rids):4d} log concentration variables to add')
        self.model.add_reactions(df_add_rids)

    def _split_and_couple_reactions(self):
        """Split and couple reactions with TD information

        reactions with TD information get split into forward / reverse
        reactions are connected to forward/reverse coupling constraints
        """
        # fbc_max_flux_val = self.model.fbc_flux_range[1]
        # fbc_max_flux_pid = self.model.get_fbc_bnd_pid(fbc_max_flux_val, 'mmol_per_gDW', 'max_reverse_flux')

        count = 0
        modify_attrs = []
        for rid, td_r in self.td_reactions.items():
            ridx = re.sub(f'^{pf.R_}', '', rid)
            modify_attrs.append([rid, 'reaction', 'product', f'{pf.C_FFC_}{ridx}=1.0'])
            if td_r.r.reversible:
                r = self.model.reactions[rid]
                rev_r = self.model.split_reversible_reaction(r)
                modify_attrs.append([rev_r.id, 'reaction', 'product', f'{pf.C_FRC_}{ridx}=1.0'])
            count += 1
        print(f'{count:4d} TD reactions split in forward/reverse')

        cols = ['id', 'component', 'attribute', 'value']
        df_modify_attrs = pd.DataFrame(modify_attrs, columns=cols)
        df_modify_attrs.set_index('id', inplace=True)
        print(f'{len(df_modify_attrs):4d} fwd/rev reactions to couple with flux direction')
        self.model.modify_attributes(df_modify_attrs, 'reaction')

    def _add_rhs_variables(self):
        """Add (fixed) RHS variables so RHS can be left at zero:
        """
        one_mmol_pid = self.model.get_fbc_bnd_pid(1.0, self.model.flux_uid, 'one_mmol', reuse=False)
        one_kj_pid = self.model.get_fbc_bnd_pid(1.0, 'kJ_per_mol', 'one_kJ', reuse=False)

        rhs_vars = {}
        rhs_flux_reactants = {sid: 1.0 for sid in self.model.species if re.match(pf.C_SU_, sid)}
        rhs_energy_reactants = {sid: 999.99 for sid in self.model.species
                                if re.match(pf.C_GFC_, sid) or re.match(pf.C_GRC_, sid)}
        rhs_vars[pf.V_RHS_FC] = [f'RHS for TD C_SU_ constraints', rhs_flux_reactants, {}, False,
                                 one_mmol_pid, one_mmol_pid, 'fixed variable', 'continuous']
        rhs_vars[pf.V_RHS_GC] = [f'RHS for TD C_GxC_ constraints', rhs_energy_reactants, {}, False,
                                 one_kj_pid, one_kj_pid, 'fixed variable', 'continuous']
        cols = ['name', 'reactants', 'products', 'reversible', 'fbcLowerFluxBound', 'fbcUpperFluxBound',
                'kind', 'notes']
        df_add_rids = pd.DataFrame(rhs_vars.values(), index=list(rhs_vars), columns=cols)
        print(f"{len(rhs_vars):4d} RHS variables to add")
        self.model.add_reactions(df_add_rids)

    def _add_tfa_variables(self):
        """Add TFA related variables to the model as pseudo reactions.

        based on pyTFA
        TFA specific variables are created and connected to constraints. Variables include:
        - [standard] transformed Gibbs energy of reaction variables ∆rG'˚ and ∆rG'
        - forward/reverse use binary variables for flux direction / ∆rG' coupling
        - log concentration variables for reactants
        - split reactions with TD info in fwd/rev and connect to coupling constraints
        - right-hand-side variables with fixed values to handle RHS not equal zero
        """
        self._add_gibbs_reaction_variables()
        self._add_use_variables()
        self._add_log_concentration_variables()
        self._split_and_couple_reactions()
        self._add_rhs_variables()

        print(f'{len(self.model.parameters):4d} parameters')

    # SLACK MODEL CONFIGURATION / TFA MODEL RELAXATION
    def _add_slack_variables(self):
        """Add slack variables for ∆rG'˚ relaxation.

        Negative and positive Slack variables are introduced according
        to pyTFA relaxation_drgo() method.
        Negative slack variables consume ∆rG' constraint (C_DRG_<rid>),
        positive slack variables produce ∆rG' constraint.
        Both slack variables are configured as non-negative (lower bound = 0)

        Following FBA optimization of the slack model (minimize sum of slack variables),
        the values of the slack variables can be used to adjust related ∆rG'˚ variable bounds.

        Slack objective and slack variables need to be removed for final TFA model
        """
        var2name = {'NS': "negative slack variable on ∆rG'˚",
                    'PS': "positive slack variable on ∆rG˚'"}

        lb_pid = self.model.get_fbc_bnd_pid(0.0, 'kJ_per_mol', 'slack_lb', reuse=False)
        ub_pid = self.model.get_fbc_bnd_pid(MAX_DRG, 'kJ_per_mol', 'slack_ub', reuse=False)

        pseudo_rids = {}
        for rid in self.td_reactions:
            ridx = re.sub(f'^{pf.R_}', '', rid)
            pseudo_rids[pf.V_NS_ + ridx] = [f'{var2name["NS"]} for {ridx}', f'{pf.C_DRG_}{ridx} =>',
                                            lb_pid, ub_pid, 'td_variable', 'continuous']
            pseudo_rids[pf.V_PS_ + ridx] = [f'{var2name["PS"]} for {ridx}', f'=> {pf.C_DRG_}{ridx}',
                                            lb_pid, ub_pid, 'td_variable', 'continuous']

        cols = ['name', 'reactionString', 'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pseudo_rids.values(), index=list(pseudo_rids), columns=cols)
        print(f"{len(df_add_rids):4d} slack variables for ∆rG'˙ to add")
        self.model.add_reactions(df_add_rids)

    def _add_slack_objective(self):
        """Add the slack objective (minimize sum over slack variables).

        Active optimization objective of model is inactivated
        The new slack objective is created and added to the model.
        """
        # inactivate active optimization objective
        for obj_id, obj in self.model.objectives.items():
            if obj.active is True:
                self.td_params['objective_id'] = obj_id
                obj.modify_attribute('active', False)

        # create new active slack_objective (minimize sum of slack variables)
        slack_obj_id = 'slack_objective'
        slack_obj_coefs = [rid for rid in self.model.reactions
                           if re.match(pf.V_PS_, rid) or re.match(pf.V_NS_, rid)]
        srefs_str = '; '.join([f'reac={var_name}, coef=1.0' for var_name in slack_obj_coefs])
        slack_obj_dict = {'type': 'minimize', 'active': True, 'fluxObjectives': srefs_str}
        self.model.objectives[slack_obj_id] = FbcObjective(pd.Series(slack_obj_dict, name=slack_obj_id))

    def export_slack_model(self, fname):
        """Export a TFA model with slack variables to perform TFA parameter relaxation.

        The slack model can be used for model parameter relaxation to adjust bounds of ∆rG'˚ variables
        and create/update the sheet 'modify_drg0_bounds' in the TFA configuration file.

        Example: Create TFA model with slack variables.

        .. code-block:: python

                xba_model = XbaModel('iML1515.xml')
                xba_model.configure()

                tfa_model = TfaModel(xba_model)
                tfa_model.configure('tfa_parameters.xlsx')
                tfa_model.export_slack_model('iML1515_TFA_slack.xml')

        Subsequently, load and optimize the slack model. Determine adjustments of variable bounds, update
        the TFA configuration file and generate a TFA model with relaxed bounds.
        See tutorial related to TFA model creation.

        :param str fname: filename with extension '.xml'
        :return: success status of operation
        :rtype: bool
        """
        self._add_slack_variables()
        self._add_slack_objective()
        self.export(fname)

    def _relax_drg0_variables(self, fluxes):
        """Relax ∆rG'˚ variable bounds based on slack model minimization.

        Nonzero negative / positive slack variables are identified and
        bounds of related ∆rG'˚ variables are updated.
        E.g. upper bound update for non-zero positive slack variable is:
            - old upper bound value + slack + epsilon (1e-6)
        Note: we are updating the values of already configured parameter ids.

        A summary of updated values is returned

        :param fluxes: CobraPy fluxes from solution object
        :type fluxes: pandas.Series or dict(str, float)
        :return: summary of ∆Gr˚ bound updates performed
        :rtype: dict
        """
        eps = 0.5
        drg0_slack = {}
        for vname, flux in fluxes.items():
            if flux > 0.0 and (re.match(pf.V_NS_, vname) or re.match(pf.V_PS_, vname)):
                drg0_slack[vname] = flux

        drg0_relaxations = {}
        modify_attrs = {}
        for vname, slack in drg0_slack.items():
            if re.match(pf.V_PS_, vname):
                ridx = re.sub(f'^{pf.V_PS_}', '', vname)
            else:
                ridx = re.sub(f'^{pf.V_NS_}', '', vname)
            drg0_var = self.model.reactions[pf.V_DRG0_ + ridx]
            drg0_lb_pid = drg0_var.fbc_lower_bound
            drg0_ub_pid = drg0_var.fbc_upper_bound
            drg0_lb_val = self.model.parameters[drg0_lb_pid].value
            drg0_ub_val = self.model.parameters[drg0_ub_pid].value

            if re.match(pf.V_NS_, vname):
                drg0_new_lb_val = drg0_lb_val - slack - eps
                modify_attrs[drg0_lb_pid] = ['parameter', 'value', drg0_new_lb_val, 'relaxation']
                drg0_relaxations[ridx] = f"lower ∆rG'˚ bound from {drg0_lb_val:8.4f} to {drg0_new_lb_val:8.4f} " \
                                         f'[{drg0_new_lb_val:8.4f}, {drg0_ub_val:8.4f}]'
            else:
                drg0_new_ub_val = drg0_ub_val + slack + eps
                modify_attrs[drg0_ub_pid] = ['parameter', 'value', drg0_new_ub_val, 'relaxation']
                drg0_relaxations[ridx] = f"upper ∆rG'˚ bound from {drg0_ub_val:8.4f} to {drg0_new_ub_val:8.4f} " \
                                         f'[{drg0_lb_val:8.4f}, {drg0_new_ub_val:8.4f}]'

        df_modify_attrs = pd.DataFrame(modify_attrs, index=['component', 'attribute', 'value', 'notes']).T
        print(f"{len(df_modify_attrs):4d} ∆Gr'˚ variables need relaxation.")
        self.model.modify_attributes(df_modify_attrs, 'parameter')
        return drg0_relaxations

    def _remove_slack_configuration(self):
        """Remove the slack configuration (variables and objective)

        Remove slack objective and negative/positive slack variables.
        Reactivate the old optimization objective
        """
        del self.model.objectives['slack_objective']
        to_del = [vname for vname in self.model.reactions
                  if re.match(pf.V_PS_, vname) or re.match(pf.V_NS_, vname)]
        for vname in to_del:
            del self.model.reactions[vname]

        old_obj_id = self.td_params['objective_id']
        self.model.objectives[old_obj_id].modify_attribute('active', True)

    def integrate_min_slack(self, fba_fluxes):
        """Integrate optimization results from slack model minimization.

        Called, with the flux values of a CobraPy optimization of the slack model

        Negative and positive slacks will be used to update bounds of related
        ∆Gr˚ variables.

        :meta private:
        :param fba_fluxes: CobraPy fluxes from solution object
        :type fba_fluxes: pandas.Series or dict(str, float)
        :return: summary of ∆Gr˚ bound updates performed
        :rtype: dict
        """
        drgo_relaxations = self._relax_drg0_variables(fba_fluxes)
        self._remove_slack_configuration()
        return drgo_relaxations

    def _modify_drg0_bounds(self, df_modify_drg0_bounds, remove_slack=True):
        """Modify ∆rG'˚ bounds after model relaxation.

        We reused existing parameter ids for ∆rG'˚. These get updated with data from df_modify_drg0_bounds.
        Index: variable id (str) for drg0 variables 'V_DRG0_<rid>'
        Columns: component: set to 'reaction' / str
        attribute: either 'fbc_lower_bound' or 'fbc_upper_bound' / str
        value: new bound value / float

        :param pandas.DataFrame df_modify_drg0_bounds: data on bounds to be updated
        :param bool remove_slack: (optional) if False do not remove slack variables (default: True)
        """
        modify_attrs = {}
        for var_id, row in df_modify_drg0_bounds.iterrows():
            drg0_var = self.model.reactions[var_id]
            specific_bound = getattr(row, 'attribute')
            pid = getattr(drg0_var, specific_bound)
            modify_attrs[pid] = ['parameter', 'value', row['value'], 'relaxation']
        cols = ['component', 'attribute', 'value', 'notes']
        df_modify_attrs = pd.DataFrame(modify_attrs.values(), index=list(modify_attrs), columns=cols)
        print(f"{len(df_modify_attrs):4d} ∆Gr'˚ variables need relaxation.")
        self.model.modify_attributes(df_modify_attrs, 'parameter')

        if remove_slack:
            self._remove_slack_configuration()



"""Implementation of TdMetabolite class.

Biochemical reactions deal with reactants, e.g. ATP
Reactants may be composed of several 'chemical' species, e.g. ATP4-, HATP3-, H2ATP2-
TD reactant has one-to-one relation with a model species
TD reactant is also linked to a single TD species with thermodynamics data

holding thermodynamic data for a TD species, like ATP4-
based on pyTFA thermo data

Reference: Book from Alberty, 2003, Thermodynamics for Biochemical Reactions

Peter Schubert, HHU Duesseldorf, Octobert 2023
"""

class TdMetabolite:

    def __init__(self, td_sid, td_sdata, all_td_cues):
        """Instantiate thermodynamic metabolite data.

        Thermodynamics data for unique metabolites
        Several model species in different compartments can link to same td_metabolite.

        Energy units in thermo data get converted, if required, to kJ/mol
        Link between model species and TD species is via td_sid

        Note: pyTFA thermo data: ∆Gf˚ at pH 0, ionic strenght of 0.0 M
            molecular structure (i.e. formula and charges), in form of
            predominant ion at pH 7, zero ionic strength and 298 K; stereochemistry is ignored
            pKa estimated using MarbinBeans (see Jankowski et al. 2008)
            iJO1366 (iAF1260) charges and formula are based on pH 7.2

        td_sdata expected to have the keys:
          'name': species name
          'formula': chemical formula
          'charged_std': electical charge in standard conditions
          'mass_std': (g/mol) molecular weight in standard conditions
          'nH_std': number of protons in standard conditions
          'deltaGf_std': Standard Gibbs energy of formation (units as given in thermo data)
          'deltaGf_err': estimation error
          'pKa': list of pKa values
          'struct_cues': dict of cues in metabolite with stoichiometry
          'error': error information (str)

        Note: for strings we convert the type from numpy.str_ to str

        :param str td_sid: id of TD species (e.g. seed id)
        :param dict all_td_cues: TD cues from TD database
        :param dict td_sdata: thermodynamics data record for metabolite extracted from TD database
        """
        self.id = str(td_sid)
        self.name = str(td_sdata['name'])
        self.formula = str(td_sdata['formula'])
        self.charge_std = td_sdata['charge_std']
        self.mass_std = td_sdata['mass_std']
        self.nh_std = td_sdata['nH_std']
        self.pkas = sorted(td_sdata['pKa'])
        self.dfg0 = td_sdata['deltaGf_std']
        self.dfg0_error = td_sdata['deltaGf_err']
        self.error = str(td_sdata['error'])
        self.struct_cues = td_sdata['struct_cues']
        self.charge_mp = 0.0
        for cue, stoic in self.struct_cues.items():
            if cue in all_td_cues:
                if all_td_cues[cue]['charge'] > 0.0:
                    self.charge_mp += float(stoic * all_td_cues[cue]['charge'])

    def modify_attribute(self, attribute, value):
        """Modify attribute value.

        :param str attribute: attribute name
        :param value: value to be configured
        :type value: str, float, int
        """
        if hasattr(self, attribute):
            setattr(self, attribute, value)
        else:
            print(f'unknown TD species attribute {attribute}')
